"""
Color playground.

Any fenced code blocks with `playground` as the language will be converted to interactive playgrounds that
are initialized with the results of the initial contents.

Any inline code blocks with `color` as the language will be evaluated as a simple color swatch.
"""
import coloraide
from coloraide import Color, NaN, Piecewise
from coloraide.interpolate import Interpolator
from pymdownx import superfences
import xml.etree.ElementTree as Etree
from collections.abc import Sequence
from collections import namedtuple
import ast
from io import StringIO
import contextlib
import sys
import re
import traceback

AST_BLOCKS = (ast.If, ast.For, ast.While, ast.Try, ast.With, ast.FunctionDef, ast.ClassDef)

RE_COLOR_START = re.compile(
    r"(?i)(?:\b(?<![-#&$])(?:color|hsla?|lch|lab|hwb|rgba?)\(|\b(?<![-#&$])[\w]{3,}(?![(-])\b|(?<![&])#)"
)

template = '''<div class="playground" id="__playground_{el_id}">
<div class="playground-results" id="__playground-results_{el_id}">
{results}
</div>
<div class="playground-code hidden" id="__playground-code_{el_id}">
<form autocomplete="off">
<textarea class="playground-inputs" id="__playground-inputs_{el_id}" spellcheck="false">{raw_source}</textarea>
</form>
</div>

<button id="__playground-edit_{el_id}" class="playground-edit" title="Edit the current snippet">Edit</button>
<button id="__playground-share_{el_id}" class="playground-share" title="Copy URL to current snippet">Share</button>
<button id="__playground-run_{el_id}" class="playground-run hidden" title="Run code (Ctrl + Enter)">Run</button>
<button id="__playground-cancel_{el_id}" class="playground-cancel hidden" title="Cancel edit (Escape)">Cancel</button>
</div>'''

code_id = 0


class ColorInterpolate(list):
    """Color interpolate."""


class ColorTuple(namedtuple('ColorTuple', ['string', 'color'])):
    """Color tuple."""


def _escape(txt):
    """Basic HTML escaping."""

    txt = txt.replace('&', '&amp;')
    txt = txt.replace('<', '&lt;')
    txt = txt.replace('>', '&gt;')
    return txt


@contextlib.contextmanager
def std_output(stdout=None):
    """Capture standard out."""
    old = sys.stdout
    if stdout is None:
        stdout = StringIO()
    sys.stdout = stdout
    yield stdout
    sys.stdout = old


def get_colors(result):
    """Get color from results."""

    colors = []
    if isinstance(result, Color):
        colors.append(ColorTuple(result.to_string(fit=False), result))
    elif isinstance(result, Interpolator):
        colors = ColorInterpolate(result.steps(steps=5, max_delta_e=4))
    elif isinstance(result, str):
        try:
            colors.append(ColorTuple(result, Color(result)))
        except Exception:
            pass
    elif isinstance(result, Sequence):
        for x in result:
            if isinstance(x, Color):
                colors.append(ColorTuple(x.to_string(fit=False), x))
            elif isinstance(x, str):
                try:
                    colors.append(ColorTuple(x, Color(x)))
                except Exception:
                    pass
    return colors


def find_colors(text):
    """Find colors in text buffer."""

    colors = []
    for m in RE_COLOR_START.finditer(text):
        start = m.start()
        mcolor = Color.match(text, start=start)
        if mcolor is not None:
            colors.append(ColorTuple(text[mcolor.start:mcolor.end], mcolor.color))
    return colors


def execute(cmd):
    """Execute color commands."""

    g = {'Color': Color, 'coloraide': coloraide, 'NaN': NaN, 'Piecewise': Piecewise}
    console = ''
    colors = []

    # Build AST tree
    src = cmd.strip()
    lines = src.split('\n')
    tree = ast.parse(src)

    for node in tree.body:
        result = None

        # Format source as Python console statements
        start = node.lineno
        end = node.end_lineno
        stmt = lines[start - 1: end]
        command = ''
        for i, line in enumerate(stmt, 0):
            if i == 0:
                stmt[i] = '>>> ' + line
            else:
                stmt[i] = '... ' + line
        command += '\n'.join(stmt)
        if isinstance(node, AST_BLOCKS):
            command += '\n... '

        # Capture anything sent to standard out
        try:
            text = ''
            with std_output() as s:
                # Execute code
                if isinstance(node, ast.Expr):
                    _eval = ast.Expression(node.value)
                    result = eval(compile(_eval, '<string>', 'eval'), g)
                else:
                    _exec = ast.Module([node], [])
                    exec(compile(_exec, '<string>', 'exec'), g)

                # Execution went well, so append command
                console += command

                # Output captured standard out after statements
                text = s.getvalue()
                if text:
                    clist = find_colors(text)
                    if clist:
                        colors.append(clist)
                    console += '\n{}'.format(text)
                s.flush()
        except Exception:
            console += '{}\n{}'.format(command, traceback.format_exc())
            # Failed for some reason, so quit
            break

        # If we got a result, output it as well
        if result is not None:
            clist = get_colors(result)
            if clist:
                colors.append(clist)
            console += '{}{}\n'.format('\n' if not text else '', str(result))
        else:
            console += '\n' if not text else ''

    return console, colors


def color_command_validator(language, inputs, options, attrs, md):
    """Color validator."""

    valid_inputs = set()

    for k, v in inputs.items():
        if k in valid_inputs:
            options[k] = True
            continue
        attrs[k] = v
    return True


def color_command_formatter(src="", language="", class_name=None, options=None, md="", **kwargs):
    """Formatter wrapper."""

    global code_id

    try:
        if len(md.preprocessors['fenced_code_block'].extension.stash) == 0:
            code_id = 0

        console, colors = execute(src.strip())
        el = ''
        bar = False
        values = []
        for item in colors:
            if isinstance(item, ColorInterpolate):
                if bar:
                    el += '<div class="swatch-bar">{}</div>'.format(' '.join(values))
                    values = []
                sub_el1 = '<div class="swatch-bar"><span class="swatch swatch-gradient">{}</span></div>'
                style = "--swatch-stops: "
                stops = []
                for color in item:
                    color.fit("srgb", in_place=True)
                    stops.append(color.convert("srgb").to_string())
                if not stops:
                    stops.extend(['transparent'] * 2)
                if len(stops) == 1:
                    stops.append(stops[0])
                style += ','.join(stops)
                sub_el2 = '<span class="swatch-color" style="{}"></span>'.format(style)
                el += sub_el1.format(sub_el2)
                bar = False
            else:
                bar = True
                base_classes = "swatch"
                for color in item:
                    if not color.color.in_gamut('srgb'):
                        base_classes += " out-of-gamut"
                    color.color.fit('srgb', in_place=True)
                    srgb = color.color.convert('srgb')
                    value1 = srgb.to_string(alpha=False)
                    value2 = srgb.to_string()
                    style = "--swatch-stops: {} 50%, {} 50%".format(value1, value2)
                    title = color.string
                    classes = base_classes
                    c = '<span class="swatch-color" style="{style}"></span>'.format(style=style)
                    c = '<span class="{classes}" title="{title}">{color}</span>'.format(
                        classes=classes,
                        color=c,
                        title=title
                    )
                    values.append(c)
        if bar:
            el += '<div class="swatch-bar">{}</div>'.format(' '.join(values))
            values = []

        el += md.preprocessors['fenced_code_block'].extension.superfences[0]['formatter'](
            src=console,
            class_name="highlight",
            language='pycon3',
            md=md,
            options=options,
            **kwargs
        )
        el = '<div class="color-command">{}</div>'.format(el)
        el = template.format(el_id=code_id, raw_source=_escape(src), results=el)
        code_id += 1
    except Exception:
        import traceback
        print(traceback.format_exc())
        return superfences.fence_code_format(src, 'text', class_name, options, md, **kwargs)

    return el


def color_formatter(src="", language="", class_name=None, md=""):
    """Formatter wrapper."""

    try:
        result = src.strip()
        try:
            console, colors = execute(result)
            if len(colors) != 1 or len(colors[0]) != 1:
                raise ValueError('Need one color only')
            color = colors[0][0].color
            result = colors[0][0].string
        except Exception:
            color = Color(result.strip())
        el = Etree.Element('span')
        stops = []
        if not color.in_gamut("srgb"):
            color.fit("srgb", in_place=True)
            attributes = {'class': "swatch out-of-gamut", "title": result}
            sub_el = Etree.SubElement(el, 'span', attributes)
            stops.append(color.convert("srgb").to_string(hex=True, alpha=False))
            if color.alpha < 1.0:
                stops[-1] += ' 50%'
                stops.append(color.convert("srgb").to_string(hex=True) + ' 50%')
        else:
            attributes = {'class': "swatch", "title": result}
            sub_el = Etree.SubElement(el, 'span', attributes)
            stops.append(color.convert("srgb").to_string(hex=True, alpha=False))
            if color.alpha < 1.0:
                stops[-1] += ' 50%'
                stops.append(color.convert("srgb").to_string(hex=True) + ' 50%')

        if not stops:
            stops.extend(['transparent'] * 2)
        if len(stops) == 1:
            stops.append(stops[0])

        Etree.SubElement(
            sub_el,
            'span',
            {
                "class": "swatch-color",
                "style": "--swatch-stops: {};".format(','.join(stops))
            }
        )

        el.append(md.inlinePatterns['backtick'].handle_code('css-color', result))
    except Exception:
        import traceback
        print(traceback.format_exc())
        el = md.inlinePatterns['backtick'].handle_code('text', src)
    return el



"""
ICtCp class.

https://professional.dolby.com/siteassets/pdfs/ictcp_dolbywhitepaper_v071.pdf
"""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, FLG_OPT_PERCENT, Labish
from .. import util
import re
from ..util import MutableVector
from typing import cast

# All PQ Values are equivalent to defaults as stated in link below:
# https://en.wikipedia.org/wiki/High-dynamic-range_video#Perceptual_quantizer
#
# ```
# M1 = 2610 / 16384
# M1INV = 16384 / 2610
# M2 = 2523 / 32
# M2INV = 32 / 2523
# C1 = 3424 / 4096
# C2 = 2413 / 128
# C3 = 2392 / 128
# ```

# XYZ transform matrices
xyz_to_lms_m = [
    [0.359132, 0.697604, -0.03578],
    [-0.192188, 1.1003800000000001, 0.07554],
    [0.006956000000000001, 0.074916, 0.8433400000000001]
]

lms_to_xyz_mi = [
    [2.0705082034204145, -1.3267039449989098, 0.2066805790352646],
    [0.3650251372337387, 0.6804585253538307, -0.04546355870112315],
    [-0.04950397021841151, -0.0495039702184115, 1.1880952852418762]
]

# LMS to Izazbz matrices
lms_p_to_ictcp_m = [
    [0.5, 0.5, 0.0],
    [1.61376953125, -3.323486328125, 1.709716796875],
    [4.378173828125, -4.24560546875, -0.132568359375]
]

ictcp_to_lms_p_mi = [
    [0.9999999999999998, 0.008609037037932761, 0.11102962500302598],
    [0.9999999999999998, -0.008609037037932752, -0.11102962500302593],
    [0.9999999999999998, 0.5600313357106791, -0.3206271749873188]
]


def ictcp_to_xyz_d65(ictcp: MutableVector) -> MutableVector:
    """From ICtCp to XYZ."""

    # Convert to LMS prime
    pqlms = cast(MutableVector, util.dot(ictcp_to_lms_p_mi, ictcp))

    # Decode PQ LMS to LMS
    lms = util.pq_st2084_eotf(pqlms)

    # Convert back to absolute XYZ D65
    absxyz = cast(MutableVector, util.dot(lms_to_xyz_mi, lms))

    # Convert back to normal XYZ D65
    return util.absxyzd65_to_xyz_d65(absxyz)


def xyz_d65_to_ictcp(xyzd65: MutableVector) -> MutableVector:
    """From XYZ to ICtCp."""

    # Convert from XYZ D65 to an absolute XYZ D5
    absxyz = util.xyz_d65_to_absxyzd65(xyzd65)

    # Convert to LMS
    lms = cast(MutableVector, util.dot(xyz_to_lms_m, absxyz))

    # PQ encode the LMS
    pqlms = util.pq_st2084_inverse_eotf(lms)

    # Calculate Izazbz
    return cast(MutableVector, util.dot(lms_p_to_ictcp_m, pqlms))


class ICtCp(Labish, Space):
    """ICtCp class."""

    BASE = "xyz-d65"
    NAME = "ictcp"
    SERIALIZE = ("--ictcp",)
    CHANNEL_NAMES = ("i", "ct", "cp")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    BOUNDS = (
        GamutUnbound(0.0, 1.0, FLG_OPT_PERCENT),
        GamutUnbound(-0.5, 0.5),
        GamutUnbound(-0.5, 0.5)
    )

    @property
    def i(self) -> float:
        """`I` channel."""

        return self._coords[0]

    @i.setter
    def i(self, value: float) -> None:
        """Set `I` channel."""

        self._coords[0] = self._handle_input(value)

    @property
    def ct(self) -> float:
        """`Ct` axis."""

        return self._coords[1]

    @ct.setter
    def ct(self, value: float) -> None:
        """`Ct` axis."""

        self._coords[1] = self._handle_input(value)

    @property
    def cp(self) -> float:
        """`Cp` axis."""

        return self._coords[2]

    @cp.setter
    def cp(self, value: float) -> None:
        """Set `Cp` axis."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ from ICtCp."""

        return ictcp_to_xyz_d65(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ to ICtCp."""

        return xyz_d65_to_ictcp(coords)



"""Utilities."""
import math
import numbers
import warnings
from functools import wraps
from typing import Optional, Sequence, List, Union, Any, Callable, Mapping, cast, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .color import Color

Vector = Sequence[float]
Matrix = Sequence[Sequence[float]]
MutableVector = List[float]
MutableMatrix = List[List[float]]
ColorInput = Union['Color', str, Mapping[str, Any]]

NaN = float('nan')
INF = float('inf')
ACHROMATIC_THRESHOLD = 0.0005
DEF_PREC = 5
DEF_FIT_TOLERANCE = 0.000075
DEF_ALPHA = 1.0
DEF_MIX = 0.5
DEF_HUE_ADJ = "shorter"
DEF_DISTANCE_SPACE = "lab"
DEF_FIT = "oklch-chroma"
DEF_DELTA_E = "76"

ERR_MAP_MSG = """
    To add or remove items from this mapping, please subclass the
    'Color' object and replace the entire mapping by either copying
    this mapping and then altering it or by creating an entirely
    new mapping. Example:

    class MyNewClass(Color):
        {name} = {{**Color.{name}, **my_override_map}}
"""

# Many libraries use 200, but `Colorjs.io` uses 203
# The author explains why 203 was chosen:
#
#   Maximum luminance in PQ is 10,000 cd/m^2
#   Relative XYZ has Y=1 for media white
#   BT.2048 says media white Y=203 at PQ 58
#
# We will currently use 203 for now as the difference is minimal.
# If there were a significant difference, and one clearly gave
# better results, that would make the decision easier, but the
# explanation above seems sufficient for now.
YW = 203

# PQ Constants
# https://en.wikipedia.org/wiki/High-dynamic-range_video#Perceptual_quantizer
M1 = 2610 / 16384
M2 = 2523 / 32
C1 = 3424 / 4096
C2 = 2413 / 128
C3 = 2392 / 128


def xy_to_xyz(xy: Vector, Y: float = 1) -> MutableVector:
    """Convert `xyY` to `xyz`."""

    x, y = xy
    return [0, 0, 0] if y == 0 else [(x * Y) / y, Y, (1 - x - y) * Y / y]


def xyz_to_uv(xyz: Vector) -> MutableVector:
    """XYZ to UV."""

    x, y, z = xyz
    denom = (x + 15 * y + 3 * z)
    if denom != 0:
        u = (4 * x) / denom
        v = (9 * y) / denom
    else:
        u = v = 0

    return [u, v]


def uv_to_xy(uv: Vector) -> MutableVector:
    """XYZ to UV."""

    u, v = uv
    denom = (6 * u - 16 * v + 12)
    if denom != 0:
        x = (9 * u) / denom
        y = (4 * v) / denom
    else:
        x = y = 0

    return [x, y]


def xy_to_uv_1960(xy: Vector) -> MutableVector:
    """XYZ to UV."""

    x, y = xy
    denom = (12 * y - 2 * x + 3)
    if denom != 0:
        u = (4 * x) / denom
        v = (6 * y) / denom
    else:
        u = v = 0

    return [u, v]


def uv_1960_to_xy(uv: Vector) -> MutableVector:
    """XYZ to UV."""

    u, v = uv
    denom = (2 * u - 8 * v + 4)
    if denom != 0:
        x = (3 * u) / denom
        y = (2 * v) / denom
    else:
        x = y = 0

    return [x, y]


def xyz_to_xyY(xyz: Vector, white: Vector) -> MutableVector:
    """XYZ to `xyY`."""

    x, y, z = xyz
    d = x + y + z
    return [white[0], white[1], y] if d == 0 else [x / d, y / d, y]


def pq_st2084_inverse_eotf(
    values: Vector,
    c1: float = C1,
    c2: float = C2,
    c3: float = C3,
    m1: float = M1,
    m2: float = M2
) -> MutableVector:
    """Perceptual quantizer (SMPTE ST 2084) - inverse EOTF."""

    adjusted = []
    for c in values:
        c = npow(c / 10000, m1)
        r = (c1 + c2 * c) / (1 + c3 * c)
        adjusted.append(npow(r, m2))
    return adjusted


def pq_st2084_eotf(
    values: Vector,
    c1: float = C1,
    c2: float = C2,
    c3: float = C3,
    m1: float = M1,
    m2: float = M2
) -> MutableVector:
    """Perceptual quantizer (SMPTE ST 2084) - EOTF."""

    im1 = 1 / m1
    im2 = 1 / m2

    adjusted = []
    for c in values:
        c = npow(c, im2)
        r = (c - c1) / (c2 - c3 * c)
        adjusted.append(10000 * npow(r, im1))
    return adjusted


def xyz_d65_to_absxyzd65(xyzd65: Vector) -> MutableVector:
    """XYZ D65 to Absolute XYZ D65."""

    return [max(c * YW, 0) for c in xyzd65]


def absxyzd65_to_xyz_d65(absxyzd65: Vector) -> MutableVector:
    """Absolute XYZ D65 XYZ D65."""

    return [max(c / YW, 0) for c in absxyzd65]


def npow(base: float, exp: float) -> float:
    """Perform `pow` with a negative number."""

    return math.copysign(abs(base) ** exp, base)


def constrain_hue(hue: float) -> float:
    """Constrain hue to 0 - 360."""

    return hue % 360 if not is_nan(hue) else hue


def is_number(value: Any) -> bool:
    """Check if value is a number."""

    return isinstance(value, numbers.Number)


def is_nan(value: float) -> bool:
    """Check if value is "not a number"."""

    return math.isnan(value)


def no_nans(value: Vector, default: float = 0.0) -> MutableVector:
    """Ensure there are no `NaN` values in a sequence."""

    return [(default if is_nan(x) else x) for x in value]


def no_nan(value: float, default: float = 0.0) -> float:
    """Convert list of numbers or single number to valid numbers."""

    return default if is_nan(value) else value


def cmp_coords(c1: Vector, c2: Vector) -> bool:
    """Compare coordinates."""

    if len(c1) != len(c2):
        return False
    else:
        return all(map(lambda a, b: (math.isnan(a) and math.isnan(b)) or a == b, c1, c2))


def dot(a: Union[float, Vector, Matrix], b: Union[float, Vector, Matrix]) -> Union[float, MutableVector, MutableMatrix]:
    """Get dot product of simple numbers, vectors, and 2D matrices and/or numbers."""

    is_a_num = is_number(a)
    is_b_num = is_number(b)
    is_a_vec = not is_a_num and is_number(cast(Union[Vector, Matrix], a)[0])
    is_b_vec = not is_b_num and is_number(cast(Union[Vector, Matrix], b)[0])
    is_a_mat = not is_a_num and not is_a_vec
    is_b_mat = not is_b_num and not is_b_vec

    if is_a_num or is_b_num:
        # Trying to dot a number with a vector or a matrix, so just multiply
        return multiply(cast(float, a), cast(float, b))
    elif is_a_vec and is_b_vec:
        # Dot product of two vectors
        return sum([x * y for x, y in zip(cast(Vector, a), cast(Vector, b))])
    elif is_a_mat and is_b_vec:
        # Dot product of matrix and a vector
        return [sum([x * y for x, y in zip(row, cast(Vector, b))]) for row in cast(Matrix, a)]
    elif is_a_vec and is_b_mat:
        # Dot product of vector and a matrix
        return [sum([x * y for x, y in zip(cast(Vector, a), col)]) for col in zip(*cast(Matrix, b))]
    else:
        # Dot product of two matrices
        return cast(
            MutableMatrix,
            [[sum(x * y for x, y in zip(row, col)) for col in zip(*cast(Matrix, b))] for row in cast(Matrix, a)]
        )


def multiply(
    a: Union[float, Vector, Matrix],
    b: Union[float, Vector, Matrix]
) -> Union[float, MutableVector, MutableMatrix]:
    """Multiply simple numbers, vectors, and 2D matrices."""

    is_a_num = is_number(a)
    is_b_num = is_number(b)
    is_a_vec = not is_a_num and is_number(cast(Union[Vector, Matrix], a)[0])
    is_b_vec = not is_b_num and is_number(cast(Union[Vector, Matrix], b)[0])
    is_a_mat = not is_a_num and not is_a_vec
    is_b_mat = not is_b_num and not is_b_vec

    if is_a_num and is_b_num:
        # Multiply two numbers
        return cast(float, a) * cast(float, b)
    elif is_a_num and not is_b_num:
        # Multiply a number and vector/matrix
        return cast(MutableVector, [multiply(cast(float, a), i) for i in cast(Union[Vector, Matrix], b)])
    elif is_b_num and not is_a_num:
        # Multiply a vector/matrix and number
        return cast(MutableVector, [multiply(i, cast(float, b)) for i in cast(Union[Vector, Matrix], a)])
    elif is_a_vec and is_b_vec:
        # Multiply two vectors
        return cast(MutableVector, [x * y for x, y in zip(cast(Vector, a), cast(Vector, b))])
    elif is_a_mat and is_b_vec:
        # Multiply matrix and a vector
        return cast(MutableMatrix, [[x * y for x, y in zip(row, cast(Vector, b))] for row in cast(Matrix, a)])
    elif is_a_vec and is_b_mat:
        # Multiply vector and a matrix
        return cast(MutableMatrix, [[x * y for x, y in zip(row, cast(Vector, a))] for row in cast(Matrix, b)])
    else:
        # Multiply two matrices
        return cast(
            MutableMatrix,
            [[x * y for x, y in zip(ra, rb)] for ra, rb in zip(cast(Matrix, a), cast(Matrix, b))]
        )


def divide(
    a: Union[float, Vector, Matrix],
    b: Union[float, Vector, Matrix]
) -> Union[float, MutableVector, MutableMatrix]:
    """Divide simple numbers, vectors, and 2D matrices."""

    is_a_num = is_number(a)
    is_b_num = is_number(b)
    is_a_vec = not is_a_num and is_number(cast(Union[Vector, Matrix], a)[0])
    is_b_vec = not is_b_num and is_number(cast(Union[Vector, Matrix], b)[0])
    is_a_mat = not is_a_num and not is_a_vec
    is_b_mat = not is_b_num and not is_b_vec

    if is_a_num and is_b_num:
        # Divide two numbers
        return cast(float, a) / cast(float, b)
    elif is_a_num and not is_b_num:
        # Divide a number and vector/matrix
        return cast(MutableVector, [divide(cast(float, a), i) for i in cast(Union[Vector, Matrix], b)])
    elif is_b_num and not is_a_num:
        # Divide a vector/matrix and number
        return cast(MutableVector, [divide(i, cast(float, b)) for i in cast(Union[Vector, Matrix], a)])
    elif is_a_vec and is_b_vec:
        # Divide two vectors
        return cast(MutableVector, [x / y for x, y in zip(cast(Vector, a), cast(Vector, b))])
    elif is_a_mat and is_b_vec:
        # Divide matrix and a vector
        return cast(MutableMatrix, [[x / y for x, y in zip(row, cast(Vector, b))] for row in cast(Matrix, a)])
    elif is_a_vec and is_b_mat:
        # Divide vector and a matrix
        return cast(MutableMatrix, [[x / y for x, y in zip(row, cast(Vector, a))] for row in cast(Matrix, b)])
    else:
        # Divide two matrices
        return cast(
            MutableMatrix,
            [[x / y for x, y in zip(ra, rb)] for ra, rb in zip(cast(Matrix, a), cast(Matrix, b))]
        )


def diag(v: Union[Vector, Matrix], k: int = 0) -> Union[MutableVector, MutableMatrix]:
    """Create a diagonal matrix from a vector or return a vector of the diagonal of a matrix."""

    size = len(v)

    if isinstance(v[0], numbers.Number):
        m = []  # type: MutableMatrix
        # Create a diagonal matrix with the provided vector
        for i, value in enumerate(cast(Vector, v)):
            m.append(([0.0] * i) + [value] + ([0.0] * (size - i - 1)))
        return m
    else:  # pragma: no cover
        d = []  # type: MutableVector
        for r in cast(Matrix, v):
            # Check that the matrix is square
            if len(r) != size:
                raise ValueError('Matrix must be a n x n matrix')
            # Return just the specified diagonal vector
            if 0 <= k < size:
                d.append(r[k])
            k += 1
        return d


def inv(matrix: Matrix) -> MutableMatrix:
    """
    Invert the matrix.

    Derived from https://github.com/ThomIves/MatrixInverse.

    While not as performant as using `numpy`, we are often caching any
    inversion we are doing, so this keeps us from having to require all
    of `numpy` for the few hits to this we do.

    This is free and unencumbered software released into the public domain.

    Anyone is free to copy, modify, publish, use, compile, sell, or
    distribute this software, either in source code form or as a compiled
    binary, for any purpose, commercial or non-commercial, and by any
    means.

    In jurisdictions that recognize copyright laws, the author or authors
    of this software dedicate any and all copyright interest in the
    software to the public domain. We make this dedication for the benefit
    of the public at large and to the detriment of our heirs and
    successors. We intend this dedication to be an overt act of
    relinquishment in perpetuity of all present and future rights to this
    software under copyright law.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
    OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
    ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.

    For more information, please refer to <http://unlicense.org/>
    """

    size = len(matrix)
    indices = list(range(size))
    m = [list(x) for x in matrix]

    # Ensure we have a square matrix
    for r in m:
        if len(r) != size:  # pragma: no cover
            raise ValueError('Matrix must be a n x n matrix')

    # Create an identity matrix of the same size as our provided vector
    im = cast(List[List[float]], diag([1] * size))

    # Iterating through each row, we will scale each row by it's "focus diagonal".
    # Then using the scaled row, we will adjust the other rows.
    # ```
    # [[fd, 0,  0 ]
    #  [0,  fd, 0 ]
    #  [0,  0,  fd]]
    # ```
    for fd in indices:
        # We will divide each value in the row by the "focus diagonal" value.
        # If the we have a zero for the given `fd` value, we cannot invert.
        denom = m[fd][fd]
        if denom == 0:  # pragma: no cover
            raise ValueError('Matrix is not invertable')

        # We are converting the matrix to the identity and vice versa,
        # So scale the diagonal such that it will now equal 1.
        # Additionally, the same operations will be applied to the identity matrix
        # and will turn it into `m ** -1` (what we are looking for)
        fd_scalar = 1.0 / denom
        for j in indices:
            m[fd][j] *= fd_scalar
            im[fd][j] *= fd_scalar

        # Now, using the value found at the index `fd` in the remaining rows (excluding `row[fd]`),
        # Where `cr` is the current row under evaluation, subtract `row[cr][fd] * row[fd] from row[cr]`.
        for cr in indices[0:fd] + indices[fd + 1:]:
            # The scalar for the current row
            cr_scalar = m[cr][fd]

            # Scale each item in the `row[fd]` and subtract it from the current row `row[cr]`
            for j in indices:
                m[cr][j] -= cr_scalar * m[fd][j]
                im[cr][j] -= cr_scalar * im[fd][j]

    # The identify matrix is now the inverse matrix and vice versa.
    return im


def cbrt(n: float) -> float:
    """Calculate cube root."""

    return nth_root(n, 3)


def nth_root(n: float, p: float) -> float:
    """Calculate nth root."""

    if p == 0:  # pragma: no cover
        return float('inf')

    if n == 0:
        # Can't do anything with zero
        return 0

    return math.copysign(abs(n) ** (p ** -1), n)


def clamp(value: float, mn: Optional[float] = None, mx: Optional[float] = None) -> float:
    """Clamp the value to the the given minimum and maximum."""

    if mn is not None and mx is not None:
        return max(min(value, mx), mn)
    elif mn is not None:
        return max(value, mn)
    elif mx is not None:
        return min(value, mx)
    else:
        return value


def fmt_float(f: float, p: int = 0) -> str:
    """
    Set float precision and trim precision zeros.

    0: Round to whole integer
    -1: Full precision
    <positive number>: precision level
    """

    if is_nan(f):
        return "none"

    value = adjust_precision(f, p)
    string = ('{{:{}f}}'.format('.53' if p == -1 else '.' + str(p))).format(value)
    return string if value.is_integer() and p == 0 else string.rstrip('0').rstrip('.')


def fmt_percent(f: float, p: int = 0) -> str:
    """Get percent."""

    if not is_nan(f):
        value = '{}%'.format(fmt_float(f, p))
    else:
        value = 'none'
    return value


def adjust_precision(f: float, p: int = 0) -> float:
    """Adjust precision."""

    if p == -1:
        return f

    elif p == 0:
        return round_half_up(f)

    else:
        whole = int(f)
        digits = 0 if whole == 0 else int(math.log10(-whole if whole < 0 else whole)) + 1
        return round_half_up(whole if digits >= p else f, p - digits)


def round_half_up(n: float, scale: int = 0) -> float:
    """Round half up."""

    mult = 10.0 ** scale
    return math.floor(n * mult + 0.5) / mult


def deprecated(message: str, stacklevel: int = 2) -> Callable[..., Any]:  # pragma: no cover
    """
    Raise a `DeprecationWarning` when wrapped function/method is called.

    Usage:

        @deprecated("This method will be removed in version X; use Y instead.")
        def some_method()"
            pass
    """

    def _wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def _deprecated_func(*args: Any, **kwargs: Any) -> Any:
            warnings.warn(
                "'{}' is deprecated. {}".format(func.__name__, message),
                category=DeprecationWarning,
                stacklevel=stacklevel
            )
            return func(*args, **kwargs)
        return _deprecated_func
    return _wrapper


def warn_deprecated(message: str, stacklevel: int = 2) -> None:  # pragma: no cover
    """Warn deprecated."""

    warnings.warn(
        message,
        category=DeprecationWarning,
        stacklevel=stacklevel
    )



"""Test Porter Duff composition modes."""
import unittest
from coloraide import Color
from . import util


class TestPorterDuffModes(util.ColorAsserts, unittest.TestCase):
    """Test Porter Duff modes."""

    def test_clear(self):
        """Test clear."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='clear'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='clear'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='clear'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='clear'),
            Color('transparent')
        )

    def test_copy(self):
        """Test copy."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='copy'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='copy'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='copy'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='copy'),
            Color('transparent')
        )

    def test_destination(self):
        """Test destination."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='destination'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='destination'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='destination'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='destination'),
            Color('transparent')
        )

    def test_source_over(self):
        """Test source over."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='source-over'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='source-over'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='source-over'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='source-over'),
            Color('transparent')
        )

    def test_destination_over(self):
        """Test destination over."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='destination-over'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='destination-over'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='destination-over'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='destination-over'),
            Color('transparent')
        )

    def test_source_in(self):
        """Test source in."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='source-in'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='source-in'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='source-in'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='source-in'),
            Color('transparent')
        )

    def test_destination_in(self):
        """Test destination in."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='destination-in'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='destination-in'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='destination-in'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='destination-in'),
            Color('transparent')
        )

    def test_source_out(self):
        """Test source out."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='source-out'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='source-out'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='source-out'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='source-out'),
            Color('transparent')
        )

    def test_destination_out(self):
        """Test destination out."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='destination-out'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='destination-out'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='destination-out'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='destination-out'),
            Color('transparent')
        )

    def test_source_atop(self):
        """Test source atop."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='source-atop'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='source-atop'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='source-atop'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='source-atop'),
            Color('transparent')
        )

    def test_destination_atop(self):
        """Test destination atop."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='destination-atop'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='destination-atop'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='destination-atop'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='destination-atop'),
            Color('transparent')
        )

    def test_xor(self):
        """Test XOR."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='xor'),
            Color('transparent')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='xor'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='xor'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='xor'),
            Color('transparent')
        )

    def test_lighter(self):
        """Test lighter."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('color(srgb 0 0 1 / 1)', operator='lighter'),
            Color('rgb(255 0 255)')
        )
        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 1)').compose('transparent', operator='lighter'),
            Color('color(srgb 1 0 0 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('color(srgb 0 0 1 / 1)', operator='lighter'),
            Color('color(srgb 0 0 1 / 1)')
        )
        self.assertColorEqual(
            Color('transparent').compose('transparent', operator='lighter'),
            Color('transparent')
        )



"""Convert the color."""
from . import util
from .util import Vector
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .color import Color

# XYZ is the absolute base, meaning that XYZ is the final base in any conversion chain.
# This is a design expectation regardless of whether someone assigns a different base to XYZ or not.
ABSOLUTE_BASE = 'xyz-d65'


def convert(color: 'Color', space: str) -> Vector:
    """Convert the color coordinates to the specified space."""

    if color.space() != space:
        obj = color.CS_MAP.get(space)
        if obj is None:
            raise ValueError("'{}' is not a valid color space".format(space))

        # Create a worse case conversion chain from XYZ to the target
        temp = obj
        count = 0
        from_color = []
        from_color_index = {}
        name = ''
        while name != ABSOLUTE_BASE:
            from_color.append(temp)
            name = temp.NAME
            from_color_index[name] = count
            temp = color.CS_MAP[temp.BASE]

            count += 1
            if count > color._MAX_CONVERT_ITERATIONS:  # pragma: no cover
                raise RuntimeError(
                    'Conversion chain reached max size of {} and has terminated to avoid an infinite loop'.format(
                        count
                    )
                )

        # Start converting coordinates until we either match a space in the conversion chain or bottom out at XYZ
        coords = util.no_nans(color.coords())
        current = type(color._space)
        if current.NAME != ABSOLUTE_BASE:
            count = 0
            while current.NAME not in from_color_index:
                # Convert to color's base
                base_space = color.CS_MAP[current.BASE]
                coords = current.to_base(coords)

                # Convert to XYZ, make sure we chromatically adapt to the appropriate white point
                if base_space.NAME == ABSOLUTE_BASE:
                    coords = color.chromatic_adaptation(current.WHITE, base_space.WHITE, coords)

                # Get next color in the chain
                current = base_space

                count += 1
                if count > color._MAX_CONVERT_ITERATIONS:  # pragma: no cover
                    raise RuntimeError(
                        'Conversions reached max iteration of {} and has terminated to avoid an infinite loop'.format(
                            count
                        )
                    )

        # If we still do not match start converting from the point in the conversion chain
        # where are current color resides
        if current.NAME != space:
            start = from_color_index[current.NAME] - 1

            # Convert from XYZ, make sure we chromatically adapt from the appropriate white point
            if current.NAME == ABSOLUTE_BASE:
                coords = color.chromatic_adaptation(current.WHITE, from_color[start].WHITE, coords)

            for index in range(start, -1, -1):
                current = from_color[index]
                coords = current.from_base(coords)

    else:
        # Nothing to convert, just pass values as is
        coords = color.coords()

    return coords



"""
Luv class.

https://en.wikipedia.org/wiki/CIELUV
"""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, FLG_PERCENT, WHITES, Labish
from .lab import KAPPA, EPSILON, KE
from .. import util
import re
from ..util import MutableVector


def xyz_to_luv(xyz: MutableVector, white: str) -> MutableVector:
    """XYZ to Luv."""

    u, v = util.xyz_to_uv(xyz)
    w_xyz = util.xy_to_xyz(WHITES[white])
    ur, vr = util.xyz_to_uv(w_xyz)

    yr = xyz[1] / w_xyz[1]
    l = 116 * util.nth_root(yr, 3) - 16 if yr > EPSILON else KAPPA * yr

    return [
        l,
        13 * l * (u - ur),
        13 * l * (v - vr),
    ]


def luv_to_xyz(luv: MutableVector, white: str) -> MutableVector:
    """Luv to XYZ."""

    l, u, v = luv
    xyz = util.xy_to_xyz(WHITES[white])
    ur, vr = util.xyz_to_uv(xyz)

    if l != 0:
        up = (u / (13 * l)) + ur
        vp = (v / (13 * l)) + vr
    else:
        up = vp = 0

    y = xyz[1] * (((l + 16) / 116) ** 3 if l > KE else l / KAPPA)

    if vp != 0:
        x = y * ((9 * up) / (4 * vp))
        z = y * ((12 - 3 * up - 20 * vp) / (4 * vp))
    else:
        x = z = 0

    return [x, y, z]


class Luv(Labish, Space):
    """Luv class."""

    BASE = "xyz-d50"
    NAME = "luv"
    SERIALIZE = ("--luv",)
    CHANNEL_NAMES = ("l", "u", "v")
    CHANNEL_ALIASES = {
        "lightness": "l"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D50"

    BOUNDS = (
        GamutUnbound(0.0, 100.0, FLG_PERCENT),
        GamutUnbound(-175.0, 175.0),
        GamutUnbound(-175.0, 175.0)
    )

    @property
    def l(self) -> float:
        """L channel."""

        return self._coords[0]

    @l.setter
    def l(self, value: float) -> None:
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def u(self) -> float:
        """U channel."""

        return self._coords[1]

    @u.setter
    def u(self, value: float) -> None:
        """U axis."""

        self._coords[1] = self._handle_input(value)

    @property
    def v(self) -> float:
        """V channel."""

        return self._coords[2]

    @v.setter
    def v(self, value: float) -> None:
        """V axis."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ D50 from Luv."""

        return luv_to_xyz(coords, cls.WHITE)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ D50 to Luv."""

        return xyz_to_luv(coords, cls.WHITE)



"""Test HSL library."""
from coloraide.util import NaN
import unittest
import math
from . import util
from coloraide import Color


class TestHSLInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test HSL."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("hsl", [20, 1, 0.75]), Color('hsl(20 100% 75%)'))

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("hsl", [20, 1, 0.75])), Color('hsl(20 100% 75%)'))

    def test_color(self):
        """Test color input/output format."""

        args = {"color": True}
        color = "color(--hsl 20 100% 75%)"

        self.assertEqual(Color(color).to_string(**args), 'color(--hsl 20 1 0.75)')

        color = "color(--hsl 20 1 0.75 / 0.5)"
        self.assertEqual(Color(color).to_string(**args), 'color(--hsl 20 1 0.75 / 0.5)')

        color = "color(--hsl 20 100% 75% / 50%)"
        self.assertEqual(Color(color).to_string(**args), 'color(--hsl 20 1 0.75 / 0.5)')

    def test_comma(self):
        """Test comma input and comma output format."""

        args = {"comma": True}

        color = "hsl(20, 100%, 75%)"
        hsl = Color(color)
        self.assertEqual(color, hsl.to_string(**args))

        color = "hsl(20, 100%, 75%, 1)"
        hsl = Color(color)
        self.assertEqual("hsl(20, 100%, 75%)", hsl.to_string(**args))

        color = "hsl(20, 100%, 75%, 0.2)"
        hsl = Color(color)
        self.assertEqual("hsla(20, 100%, 75%, 0.2)", hsl.to_string(**args))

    def test_space(self):
        """Test space input and space output format."""

        args = {}

        color = "hsl(20 100% 75%)"
        hsl = Color(color)
        self.assertEqual(color, hsl.to_string(**args))

        color = "hsl(20 100% 75% / 1)"
        hsl = Color(color)
        self.assertEqual("hsl(20 100% 75%)", hsl.to_string(**args))

        color = "hsl(20 100% 75% / 0.2)"
        hsl = Color(color)
        self.assertEqual(color, hsl.to_string(**args))

    def test_alias(self):
        """Test that `hsla` is an alias for `hsl`."""

        args = {"comma": True}

        color = "hsla(20, 100%, 75%)"
        hsl = Color(color)
        self.assertEqual("hsl(20, 100%, 75%)", hsl.to_string(**args))

        color = "hsla(20, 100%, 75%, 1)"
        hsl = Color(color)
        self.assertEqual("hsl(20, 100%, 75%)", hsl.to_string(**args))

        color = "hsla(20, 100%, 75%, 0.5)"
        hsl = Color(color)
        self.assertEqual("hsla(20, 100%, 75%, 0.5)", hsl.to_string(**args))

        args["comma"] = False

        color = "hsla(20, 100%, 75%)"
        hsl = Color(color)
        self.assertEqual("hsl(20 100% 75%)", hsl.to_string(**args))

        color = "hsla(20, 100%, 75%, 1)"
        hsl = Color(color)
        self.assertEqual("hsl(20 100% 75%)", hsl.to_string(**args))

        color = "hsla(20, 100%, 75%, 0.5)"
        hsl = Color(color)
        self.assertEqual("hsl(20 100% 75% / 0.5)", hsl.to_string(**args))

    def test_percent(self):
        """Test that percents work properly."""

        args = {"comma": True}

        color = "hsla(20, 100%, 75%, 100%)"
        hsl = Color(color)
        self.assertEqual("hsl(20, 100%, 75%)", hsl.to_string(**args))

        color = "hsla(20, 100%, 75%, 20%)"
        hsl = Color(color)
        self.assertEqual("hsla(20, 100%, 75%, 0.2)", hsl.to_string(**args))

        args["comma"] = False

        color = "hsl(20 100% 75% / 100%)"
        hsl = Color(color)
        self.assertEqual("hsl(20 100% 75%)", hsl.to_string(**args))

        color = "hsl(20 100% 75% / 20%)"
        hsl = Color(color)
        self.assertEqual("hsl(20 100% 75% / 0.2)", hsl.to_string(**args))

    def test_no_alpha(self):
        """Test no alpha."""

        args = {"comma": True, "alpha": False}

        color = "hsla(20, 100%, 75%, 0.2)"
        hsl = Color(color)
        self.assertEqual("hsl(20, 100%, 75%)", hsl.to_string(**args))

        args["comma"] = False

        color = "hsl(20 100% 75% / 0.2)"
        hsl = Color(color)
        self.assertEqual("hsl(20 100% 75%)", hsl.to_string(**args))

    def test_force_alpha(self):
        """Test force alpha."""

        args = {"comma": True, "alpha": True}

        color = "hsla(20, 100%, 75%, 1)"
        hsl = Color(color)
        self.assertEqual("hsla(20, 100%, 75%, 1)", hsl.to_string(**args))

        args["comma"] = False

        color = "hsl(20 100% 75% / 1)"
        hsl = Color(color)
        self.assertEqual("hsl(20 100% 75% / 1)", hsl.to_string(**args))

    def test_precision(self):
        """Test precision."""

        color = 'color(--hsl 20.1234567 0.1234567 0.1234567)'
        self.assertEqual(Color(color).to_string(), 'hsl(20.123 12.346% 12.346%)')
        self.assertEqual(Color(color).to_string(precision=3), 'hsl(20.1 12.3% 12.3%)')
        self.assertEqual(Color(color).to_string(precision=0), 'hsl(20 12% 12%)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'hsl(20.12345669999999842048055143095552921295166015625 12.3456700000000001438138497178442776203155517578125% 12.3456700000000001438138497178442776203155517578125%)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('color(--hsl 20 150% 75%)').to_string(),
            'hsl(19.319 100% 77.154%)'
        )

        self.assertEqual(
            Color('color(--hsl 20 150% 75%)').to_string(fit="clip"),
            'hsl(20 100% 75%)'
        )

        self.assertEqual(
            Color('color(--hsl 20 150% 75%)').to_string(fit=False),
            'hsl(20 150% 75%)'
        )

    def test_hue_inputs(self):
        """Test hue inputs."""

        color = "hsl(90deg, 100%, 75%)"
        hsl = Color(color)
        self.assertEqual("hsl(90 100% 75%)", hsl.to_string())

        color = "hsl({:f}rad, 100%, 75%)".format(math.radians(90))
        hsl = Color(color)
        self.assertEqual("hsl(90 100% 75%)", hsl.to_string())

        color = "hsl(100grad, 100%, 75%)"
        hsl = Color(color)
        self.assertEqual("hsl(90 100% 75%)", hsl.to_string())

        color = "hsl(0.25turn, 100%, 75%)"
        hsl = Color(color)
        self.assertEqual("hsl(90 100% 75%)", hsl.to_string())


class TestHSLProperties(util.ColorAsserts, unittest.TestCase):
    """Test HSL."""

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--hsl 120 50% 90% / 1)')
        self.assertEqual(c.hue, 120)
        c.hue = 110
        self.assertEqual(c.hue, 110)

    def test_saturation(self):
        """Test `saturation`."""

        c = Color('color(--hsl 120 50% 90% / 1)')
        self.assertEqual(c.saturation, 0.5)
        c.saturation = 0.6
        self.assertEqual(c.saturation, 0.6)

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--hsl 120 50% 90% / 1)')
        self.assertEqual(c.lightness, 0.9)
        c.lightness = 0.8
        self.assertEqual(c.lightness, 0.8)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--hsl 120 50% 90% / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('hsl', [NaN, 0.5, 1], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_auto_null(self):
        """Test auto null."""

        c = Color('hsl(120 0% 75% / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_corner_case_null(self):
        """Test corner case that produces null."""

        c = Color('color(srgb -2 0 2)').convert('hsl')
        self.assertTrue(c.is_nan('hue'))



"""Test utilities."""
import unittest
import decimal
from coloraide import util, NaN


class TestUtil(unittest.TestCase):
    """Test utilities."""

    def test_is_number(self):
        """Test if is a number."""

        self.assertTrue(util.is_number(3))
        self.assertTrue(util.is_number(3.3))
        self.assertTrue(util.is_number(decimal.Decimal('3.3')))
        self.assertFalse(util.is_number('3.3'))

    def test_no_nan(self):
        """Test no `NaN`."""

        self.assertEqual(util.no_nan(NaN), 0)
        self.assertEqual(util.no_nans([0, 1, 2, NaN]), [0, 1, 2, 0])

    def test_compare_coords(self):
        """Test coordinate comparison."""

        self.assertTrue(util.cmp_coords([1, 2, 3], [1, 2, 3]))
        self.assertTrue(util.cmp_coords([1, NaN, 3], [1, NaN, 3]))
        self.assertFalse(util.cmp_coords([1, 3, 2], [1, 2, 3]))
        self.assertFalse(util.cmp_coords([1, 2, NaN], [1, NaN, 3]))
        self.assertFalse(util.cmp_coords([1, 2, 3, 4], [1, 2, 3]))

    def test_is_nan(self):
        """Test if is `NaN`."""

        self.assertTrue(util.is_nan(float('nan')))
        self.assertTrue(NaN)
        self.assertFalse(util.is_nan(3))
        self.assertFalse(util.is_nan(0))

    def test_round(self):
        """Test rounding."""

        self.assertEqual(util.round_half_up(3.3), 3)
        self.assertEqual(util.round_half_up(3.5), 4)
        self.assertEqual(util.round_half_up(3.9), 4)
        self.assertEqual(util.round_half_up(4), 4)

    def test_scale(self):
        """Test rounding."""

        self.assertEqual(util.round_half_up(3.345, 1), 3.3)
        self.assertEqual(util.round_half_up(3.345, 2), 3.35)
        self.assertEqual(util.round_half_up(3.345, 3), 3.345)
        self.assertEqual(util.round_half_up(333, -2), 300)

    def test_cbrt(self):
        """Test cube root."""

        self.assertEqual(util.cbrt(27), 3)
        self.assertEqual(util.cbrt(-27), -3)

    def test_clamp(self):
        """Test clamp."""

        self.assertEqual(util.clamp(3, None, None), 3)
        self.assertEqual(util.clamp(3, 4, None), 4)
        self.assertEqual(util.clamp(4, 4, None), 4)
        self.assertEqual(util.clamp(4, None, 4), 4)
        self.assertEqual(util.clamp(5, None, 4), 4)
        self.assertEqual(util.clamp(3, 4, 6), 4)
        self.assertEqual(util.clamp(7, 4, 6), 6)
        self.assertEqual(util.clamp(4, 4, 6), 4)
        self.assertEqual(util.clamp(6, 4, 6), 6)

    def test_format_float(self):
        """Test formatting of floats."""

        self.assertEqual(util.fmt_float(0.123456, 5), "0.12346")
        self.assertEqual(util.fmt_float(0.123456, 4), "0.1235")
        self.assertEqual(util.fmt_float(2.123456, 0), "2")
        self.assertEqual(util.fmt_float(2.123456, -1), "2.12345600000000001017497197608463466167449951171875")

    def test_dot(self):
        """Test dot."""

        self.assertEqual(util.dot(2, 2), 4)
        self.assertEqual(util.dot([1, 2, 3], 2), [2, 4, 6])
        self.assertEqual(util.dot(2, [1, 2, 3]), [2, 4, 6])
        self.assertEqual(util.dot([1, 2, 3], [4, 5, 6]), 32)
        self.assertEqual(util.dot([4, 5, 6], [1, 2, 3]), 32)
        self.assertEqual(
            util.dot(
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                [1, 2, 3]
            ),
            [14, 32, 50]
        )
        self.assertEqual(
            util.dot(
                [1, 2, 3],
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            ),
            [30, 36, 42]
        )
        self.assertEqual(
            util.dot(
                [[4, 4, 4], [1, 0, 1], [2, 3, 4]],
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            ),
            [[48, 60, 72], [8, 10, 12], [42, 51, 60]]
        )

    def test_multiply(self):
        """Test multiply."""

        self.assertEqual(util.multiply(2, 2), 4)
        self.assertEqual(util.multiply([1, 2, 3], 2), [2, 4, 6])
        self.assertEqual(util.multiply(2, [1, 2, 3]), [2, 4, 6])
        self.assertEqual(util.multiply([1, 2, 3], [4, 5, 6]), [4, 10, 18])
        self.assertEqual(util.multiply([4, 5, 6], [1, 2, 3]), [4, 10, 18])
        self.assertEqual(
            util.multiply(
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                [1, 2, 3]
            ),
            [[1, 4, 9], [4, 10, 18], [7, 16, 27]]
        )
        self.assertEqual(
            util.multiply(
                [1, 2, 3],
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            ),
            [[1, 4, 9], [4, 10, 18], [7, 16, 27]]
        )
        self.assertEqual(
            util.multiply(
                [[4, 4, 4], [1, 0, 1], [2, 3, 4]],
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            ),
            [[4, 8, 12], [4, 0, 6], [14, 24, 36]]
        )

    def test_divide(self):
        """Test divide."""

        self.assertEqual(util.divide(4, 2), 2)
        self.assertEqual(util.divide([2, 4, 6], 2), [1, 2, 3])
        self.assertEqual(util.divide(2, [2, 4, 6]), [1.0, 0.5, 0.3333333333333333])
        self.assertEqual(util.divide([4, 10, 18], [4, 5, 6]), [1, 2, 3])
        self.assertEqual(util.divide([4, 10, 18], [1, 2, 3]), [4, 5, 6])
        self.assertEqual(
            util.divide(
                [[1, 4, 9], [4, 10, 18], [7, 16, 27]],
                [1, 2, 3]
            ),
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        )
        self.assertEqual(
            util.divide(
                [1, 2, 3],
                [[1, 4, 9], [4, 10, 18], [7, 16, 27]]
            ),
            [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        )
        self.assertEqual(
            util.divide(
                [[1, 4, 9], [4, 10, 18], [7, 16, 27]],
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            ),
            [[1.0, 2.0, 3.0], [1.0, 2.0, 3.0], [1.0, 2.0, 3.0]]
        )
        self.assertEqual(
            util.divide(
                [[4, 8, 12], [4, 0, 6], [14, 24, 36]],
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            ),
            [[4, 4, 4], [1, 0, 1], [2, 3, 4]]
        )



"""Rec 2020 color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb import SRGB
from .. import util
import re
import math
from ..util import MutableVector
from typing import cast

ALPHA = 1.09929682680944
BETA = 0.018053968510807
BETA45 = 0.018053968510807 * 4.5

RGB_TO_XYZ = [
    [6.3695804830129132e-01, 1.4461690358620838e-01, 1.6888097516417216e-01],
    [2.6270021201126703e-01, 6.7799807151887104e-01, 5.9301716469861973e-02],
    [4.9941065744660755e-17, 2.8072693049087438e-02, 1.0609850577107913e+00]
]

XYZ_TO_RGB = [
    [1.7166511879712676, -0.3556707837763924, -0.2533662813736598],
    [-0.6666843518324889, 1.6164812366349388, 0.01576854581391115],
    [0.01763985744531078, -0.04277061325780851, 0.9421031212354736]
]


def lin_2020(rgb: MutableVector) -> MutableVector:
    """
    Convert an array of rec-2020 RGB values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    https://en.wikipedia.org/wiki/Rec._2020#Transfer_characteristics
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < BETA45:
            result.append(i / 4.5)
        else:
            result.append(math.copysign(util.nth_root((abs_i + ALPHA - 1) / ALPHA, 0.45), i))
    return result


def gam_2020(rgb: MutableVector) -> MutableVector:
    """
    Convert an array of linear-light rec-2020 RGB  in the range 0.0-1.0 to gamma corrected form.

    https://en.wikipedia.org/wiki/Rec._2020#Transfer_characteristics
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < BETA:
            result.append(4.5 * i)
        else:
            result.append(math.copysign(ALPHA * (abs_i ** 0.45) - (ALPHA - 1), i))
    return result


def lin_2020_to_xyz(rgb: MutableVector) -> MutableVector:
    """
    Convert an array of linear-light rec-2020 values to CIE XYZ using  D65.

    (no chromatic adaptation)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    return cast(MutableVector, util.dot(RGB_TO_XYZ, rgb))


def xyz_to_lin_2020(xyz: MutableVector) -> MutableVector:
    """Convert XYZ to linear-light rec-2020."""

    return cast(MutableVector, util.dot(XYZ_TO_RGB, xyz))


class Rec2020(SRGB):
    """Rec 2020 class."""

    BASE = "xyz-d65"
    NAME = "rec2020"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=NAME, channels=3))
    WHITE = "D65"

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ from Rec 2020."""

        return lin_2020_to_xyz(lin_2020(coords))

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ to Rec 2020."""

        return gam_2020(xyz_to_lin_2020(coords))



"""
Slice diagrams.

Show a slice of a color space where we plot the relation between two channels holding one constant.

Channels to be plotted are given in the form `name:distance:offset`. For instance `a:1:-0.5` would plot the `a` channel
from -0.5 to 0.5.

The constant channel is given in the form `name:value` where value is the constant value for the channel.
"""
import itertools
import matplotlib.pyplot as plt
import argparse
import sys
import os

sys.path.insert(0, os.getcwd())

from coloraide import Color, NaN  # noqa: E402


def needs_lchuv_workaround(color):
    """
    Check if LCHuv has high chroma and no lightness.

    LCHuv will have such values all in gamut and create weird graphs.
    This is just due to how LCHuv algorithm handles high chroma with and zero lightness,
    it all gets treated as black which is in gamut for almost any color.
    """

    return color.space().startswith('lchuv') and color.l == 0 and not color.normalize().is_nan('hue')


def plot_slice(space, channel0, channel1, channel2, gamut='srgb', resolution=500, dark=False, title=""):
    """Plot a slice."""

    res = resolution
    if not dark:
        plt.style.use('seaborn-whitegrid')
        default_color = 'black'
    else:
        plt.style.use('dark_background')
        default_color = 'white'

    figure = plt.figure()

    # Create a color object based on the specified space.
    c = Color(space, [])

    # Parse the channel strings into actual values
    name0, value = [c if i == 0 else float(c) for i, c in enumerate(channel0.split(":"), 0)]
    name1, factor1, offset1 = [c if i == 0 else float(c) for i, c in enumerate(channel1.split(':'), 0)]
    name2, factor2, offset2 = [c if i == 0 else float(c) for i, c in enumerate(channel2.split(':'), 0)]

    # Get the actual indexes of the specified channels
    name0 = c._space.CHANNEL_ALIASES.get(name0, name0)
    index0 = c._space.CHANNEL_NAMES.index(name0)
    name1 = c._space.CHANNEL_ALIASES.get(name1, name1)
    index1 = c._space.CHANNEL_NAMES.index(name1)
    name2 = c._space.CHANNEL_ALIASES.get(name2, name2)
    index2 = c._space.CHANNEL_NAMES.index(name2)

    # Arrays for data points to plot
    c_map = []
    xaxis = []
    yaxis = []

    # Track minimum and maximum values
    c1_mn = float('inf')
    c1_mx = float('-inf')
    c2_mn = float('inf')
    c2_mx = float('-inf')

    # Track the edge of the graphed shape.
    edge_map = {}

    # Iterate through the two specified channels
    for c1, c2 in itertools.product(
        ((x / res * factor1) + offset1 for x in range(0, res + 1)),
        ((x / res * factor2) + offset2 for x in range(0, res + 1))
    ):
        # Set the appropriate channels and update the color object
        coords = [NaN] * 3
        coords[index0] = value
        coords[index1] = c1
        coords[index2] = c2
        c.update(space, coords)

        # Only process colors within the gamut of sRGB.
        if c.in_gamut(gamut, tolerance=0) and not needs_lchuv_workaround(c):
            # Get the absolute min and max value plotted
            if c1 < c1_mn:
                c1_mn = c1
            if c1 > c1_mx:
                c1_mx = c1

            if c2 < c2_mn:
                c2_mn = c2
            if c2 > c2_mx:
                c2_mx = c2

            # Create an edge map so we can draw an outline
            if c1 not in edge_map:
                mn = mx = c2
            else:
                mn, mx = edge_map[c1]
                if c2 < mn:
                    mn = c2
                elif c2 > mx:
                    mx = c2
            edge_map[c1] = [mn, mx]

            # Save the points
            xaxis.append(c1)
            yaxis.append(c2)
            c_map.append(c.convert('srgb').to_string(hex=True))

    # Create a border around the data
    xe = []
    ye = []
    xtemp = []
    ytemp = []
    for p1, edges in edge_map.items():
        xe.append(p1)
        ye.append(edges[0])
        xtemp.append(p1)
        ytemp.append(edges[1])
    xe.extend(reversed(xtemp))
    ye.extend(reversed(ytemp))
    xe.append(xe[0])
    ye.append(ye[0])

    ax = plt.axes(
        xlabel='{}: {} - {}'.format(name1, c1_mn, c1_mx),
        ylabel='{}: {} - {}'.format(name2, c2_mn, c2_mx)
    )
    ax.set_aspect('auto')
    figure.add_axes(ax)

    if not title:
        title = "A Slice of '{}' and '{}' in the {} Color Space".format(name1, name2, space)
    title += '\n{}: {}'.format(name0, value)
    plt.title(title)

    plt.plot(xe, ye, color=default_color, marker="", linewidth=2, markersize=0, antialiased=True)

    plt.scatter(
        xaxis,
        yaxis,
        marker="o",
        color=c_map,
        s=2
    )


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='slice_diagrams', description='Plot a slice of a color space.')
    parser.add_argument('--space', '-s', help='Desired space.')
    parser.add_argument('--gamut', '-g', default="srgb", help='Gamut to evaluate the color in (default is sRGB).')
    parser.add_argument('--constant', '-c', help="The channel to hold constant and the value to use 'name:value'.")
    parser.add_argument('--xaxis', '-x', help="The channel to plot on X axis 'name:range:offset'.")
    parser.add_argument('--yaxis', '-y', help="The channel to plot on Y axis 'name:range:offset'.")
    parser.add_argument('--resolution', '-r', default="800", help="How densely to render the figure.")
    parser.add_argument('--title', '-t', default='', help="Provide a title for the diagram.")
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--output', '-o', default='', help='Output file.')

    args = parser.parse_args()

    plot_slice(
        args.space,
        args.constant,
        args.xaxis,
        args.yaxis,
        gamut=args.gamut,
        resolution=int(args.resolution),
        title=args.title,
        dark=args.dark
    )

    if args.output:
        plt.savefig(args.output)
    else:
        plt.show()


if __name__ == "__main__":
    sys.exit(main())



"""Din99o Lch class."""
from ..spaces import RE_DEFAULT_MATCH
from .lch import Lch
from .. import util
import math
import re
from ..util import MutableVector

ACHROMATIC_THRESHOLD = 0.0000000002


def lch_to_lab(lch: MutableVector) -> MutableVector:
    """Din99o Lch to lab."""

    l, c, h = lch
    h = util.no_nan(h)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if c < 0.0:
        c = 0.0

    return [
        l,
        c * math.cos(math.radians(h)),
        c * math.sin(math.radians(h))
    ]


def lab_to_lch(lab: MutableVector) -> MutableVector:
    """Din99o Lab to Lch."""

    l, a, b = lab
    h = math.degrees(math.atan2(b, a))
    c = math.sqrt(a ** 2 + b ** 2)

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if c <= ACHROMATIC_THRESHOLD:
        h = util.NaN

    return [l, c, util.constrain_hue(h)]


class Din99oLch(Lch):
    """Din99o Lch class."""

    BASE = 'din99o'
    NAME = "din99o-lch"
    SERIALIZE = ("--din99o-lch",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To Din99o from Din99o Lch."""

        return lch_to_lab(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From Din99o to Din99o Lch."""

        return lab_to_lch(coords)



"""Test Okhsv library."""
import unittest
from . import util
from coloraide import Color, NaN


class TestOkhsvInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test Okhsv."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("okhsv", [20, 1, 0.75]), Color('color(--okhsv 20 100% 75%)'))

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("okhsv", [20, 1, 0.75])), Color('color(--okhsv 20 100% 75%)'))

    def test_color(self):
        """Test color input/output format."""

        args = {"color": True}
        color = "color(--okhsv 20 100% 75%)"

        self.assertEqual(Color(color).to_string(**args), 'color(--okhsv 20 1 0.75)')

        color = "color(--okhsv 20 1 0.75 / 0.5)"
        self.assertEqual(Color(color).to_string(**args), 'color(--okhsv 20 1 0.75 / 0.5)')

        color = "color(--okhsv 20 100% 75% / 50%)"
        self.assertEqual(Color(color).to_string(**args), 'color(--okhsv 20 1 0.75 / 0.5)')

    def test_no_alpha(self):
        """Test no alpha."""

        args = {"alpha": False}

        color = "color(--okhsv 20 100% 75% 0.2)"
        okhsv = Color(color)
        self.assertEqual("color(--okhsv 20 1 0.75)", okhsv.to_string(**args))

    def test_force_alpha(self):
        """Test force alpha."""

        args = {"alpha": True}

        color = "color(--okhsv 20 100% 75% / 1)"
        okhsv = Color(color)
        self.assertEqual("color(--okhsv 20 1 0.75 / 1)", okhsv.to_string(**args))

    def test_precision(self):
        """Test precision."""

        color = 'color(--okhsv 20.1234567 0.1234567 0.1234567)'
        self.assertEqual(Color(color).to_string(), 'color(--okhsv 20.123 0.12346 0.12346)')
        self.assertEqual(Color(color).to_string(precision=3), 'color(--okhsv 20.1 0.123 0.123)')
        self.assertEqual(Color(color).to_string(precision=0), 'color(--okhsv 20 0 0)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'color(--okhsv 20.12345669999999842048055143095552921295166015625 0.12345670000000000254836152180359931662678718566894531 0.12345670000000000254836152180359931662678718566894531)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('color(--okhsv 20 150% 75%)').to_string(),
            'color(--okhsv 20 1 0.60577)'
        )

        self.assertEqual(
            Color('color(--okhsv 20 150% 75%)').to_string(fit="clip"),
            'color(--okhsv 20 1 0.75)'
        )

        self.assertEqual(
            Color('color(--okhsv 20 150% 75%)').to_string(fit=False),
            'color(--okhsv 20 1.5 0.75)'
        )


class TestOkhsvProperties(util.ColorAsserts, unittest.TestCase):
    """Test Okhsv."""

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--okhsv 120 50% 50% / 1)')
        self.assertEqual(c.hue, 120)
        c.hue = 110
        self.assertEqual(c.hue, 110)

    def test_saturation(self):
        """Test `saturation`."""

        c = Color('color(--okhsv 120 50% 50% / 1)')
        self.assertEqual(c.saturation, 0.5)
        c.saturation = 0.6
        self.assertEqual(c.saturation, 0.6)

    def test_value(self):
        """Test `value`."""

        c = Color('color(--okhsv 120 50% 50% / 1)')
        self.assertEqual(c.value, 0.5)
        c.value = 0.4
        self.assertEqual(c.value, 0.4)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--okhsv 120 50% 50% / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('okhsv', [NaN, 0.5, 0.75], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_auto_null(self):
        """Test auto null."""

        c = Color('color(--okhsv 120 0% 75% / 1)')
        self.assertTrue(c.is_nan('hue'))



"""
Luv class.

https://en.wikipedia.org/wiki/CIELUV
"""
from ..spaces import RE_DEFAULT_MATCH
from .luv import Luv
import re


class LuvD65(Luv):
    """Oklab class."""

    BASE = "xyz-d65"
    NAME = "luv-d65"
    SERIALIZE = ("--luv-d65",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"



"""Test Luv library."""
import unittest
from . import util
from coloraide import Color


class TestLuvInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test Luv."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("luv", [20, 10, -30]), Color('color(--luv 20% 10 -30)'))

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("luv", [20, 10, -30])), Color('color(--luv 20% 10 -30)'))

    def test_color(self):
        """Test color input/output format."""

        args = {"color": True}
        color = "color(--luv 20% 10 -30)"

        self.assertEqual(Color(color).to_string(**args), 'color(--luv 20 10 -30)')

        color = "color(--luv 20% 10 -30 / 0.5)"
        self.assertEqual(Color(color).to_string(**args), 'color(--luv 20 10 -30 / 0.5)')

        color = "color(--luv 20% 10 -30 / 50%)"
        self.assertEqual(Color(color).to_string(**args), 'color(--luv 20 10 -30 / 0.5)')

    def test_percent(self):
        """Test that percents work properly."""

        color = "color(--luv 20% 10 -30 / 100%)"
        luv = Color(color)
        self.assertEqual("color(--luv 20 10 -30)", luv.to_string())

        color = "color(--luv 20% 10 -30 / 20%)"
        luv = Color(color)
        self.assertEqual("color(--luv 20 10 -30 / 0.2)", luv.to_string())

    def test_no_alpha(self):
        """Test no alpha."""

        color = "color(--luv 20% 10 -30 / 0.2)"
        luv = Color(color)
        self.assertEqual("color(--luv 20 10 -30)", luv.to_string(alpha=False))

    def test_force_alpha(self):
        """Test force alpha."""

        color = "color(--luv 20% 10 -30 / 1)"
        luv = Color(color)
        self.assertEqual("color(--luv 20 10 -30 / 1)", luv.to_string(alpha=True))

    def test_precision(self):
        """Test precision."""

        color = 'color(--luv 20.1234567% 10.1234567 -30.1234567)'
        self.assertEqual(Color(color).to_string(), 'color(--luv 20.123 10.123 -30.123)')
        self.assertEqual(Color(color).to_string(precision=3), 'color(--luv 20.1 10.1 -30.1)')
        self.assertEqual(Color(color).to_string(precision=0), 'color(--luv 20 10 -30)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'color(--luv 20.12345669999999842048055143095552921295166015625 10.1234567000000001968373908312059938907623291015625 -30.12345669999999842048055143095552921295166015625)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('color(--luv -20% 180 -180)').to_string(),
            'color(--luv -20 180 -180)'
        )

        self.assertEqual(
            Color('color(--luv -20% 180 -180)').to_string(fit="clip"),
            'color(--luv -20 180 -180)'
        )

        self.assertEqual(
            Color('color(--luv -20% 180 -180)').to_string(fit=False),
            'color(--luv -20 180 -180)'
        )


class TestLuvProperties(util.ColorAsserts, unittest.TestCase):
    """Test Luv."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--luv 90% 50 -20 / 1)')
        self.assertEqual(c.lightness, 90)
        c.lightness = 80
        self.assertEqual(c.lightness, 80)

    def test_u(self):
        """Test `u`."""

        c = Color('color(--luv 90% 50 -20 / 1)')
        self.assertEqual(c.u, 50)
        c.u = 40
        self.assertEqual(c.u, 40)

    def test_v(self):
        """Test `v`."""

        c = Color('color(--luv 90% 50 -20 / 1)')
        self.assertEqual(c.v, -20)
        c.v = -10
        self.assertEqual(c.v, -10)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--luv 90% 50 -20 / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)



"""Fit by compressing chroma in Lch."""
from .fit_oklch_chroma import OklchChroma


class LchChroma(OklchChroma):
    """Lch chroma gamut mapping class."""

    NAME = "lch-chroma"

    EPSILON = 0.01
    LIMIT = 2.0
    DE = "2000"
    SPACE = "lch"
    SPACE_COORDINATE = "{}.chroma".format(SPACE)



"""Test Blend modes."""
import unittest
from coloraide import Color
from . import util

# Colors that produce pretty distinct results
REDISH = '#fc3d99'
BLUISH = '#07c7ed'
YELLOWISH = '#f5d311'


class TestBlendModes(util.ColorAsserts, unittest.TestCase):
    """Test blend modes."""

    def test_alpha(self):
        """Test normal blend mode with source alpha."""

        self.assertColorEqual(Color('blue').compose('color(srgb 1 0 0 / 0.5)', blend='normal'), Color('blue'))
        self.assertColorEqual(
            Color('color(srgb 0 0 1 / 0.5)').compose('color(srgb 1 0 0)', blend='normal'),
            Color('color(srgb 0.5 0 0.5)')
        )
        self.assertColorEqual(
            Color('color(srgb 0 0 1 / 0.5)').compose('color(srgb 1 0 0 / 0.5)', blend='normal'),
            Color('color(srgb 0.25 0 0.5 / 0.75)')
        )

    def test_normal(self):
        """Test normal."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='normal'), Color(REDISH))
        self.assertColorEqual(Color(REDISH).compose('white', blend='normal'), Color(REDISH))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='normal'), Color(REDISH))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='normal'), Color(BLUISH))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='normal'), Color(BLUISH))

    def test_multiply(self):
        """Test multiply."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='multiply'), Color('black'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='multiply'), Color(REDISH))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='multiply'), Color('rgb(242.12 50.475 10.2)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='multiply'), Color('rgb(6.9176 47.604 142.2)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='multiply'), Color('rgb(6.7255 164.66 15.8)'))

    def test_screen(self):
        """Test screen."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='screen'), Color(REDISH))
        self.assertColorEqual(Color(REDISH).compose('white', blend='screen'), Color('white'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='screen'), Color('rgb(254.88 221.53 159.8)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='screen'), Color('rgb(252.08 212.4 247.8)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='screen'), Color('rgb(245.27 245.34 238.2)'))

    def test_overlay(self):
        """Test overlay."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='overlay'), Color('black'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='overlay'), Color('white'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='overlay'), Color('rgb(254.76 188.05 20.4)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='overlay'), Color('rgb(249.16 95.208 240.6)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='overlay'), Color('rgb(235.55 235.67 31.6)'))

    def test_darken(self):
        """Test darken."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='darken'), Color('black'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='darken'), Color('rgb(252 61 153)'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='darken'), Color('rgb(245 61 17)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='darken'), Color('rgb(7 61 153)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='darken'), Color('rgb(7 199 17)'))

    def test_lighten(self):
        """Test lighten."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='lighten'), Color('rgb(252 61 153)'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='lighten'), Color('white'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='lighten'), Color('rgb(252 211 153)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='lighten'), Color('rgb(252 199 237)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='lighten'), Color('rgb(245 211 237)'))

    def test_color_dodge(self):
        """Test color dodge."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='color-dodge'), Color('black'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='color-dodge'), Color('white'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='color-dodge'), Color('rgb(255 255 42.5)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='color-dodge'), Color('rgb(255 255 255)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='color-dodge'), Color('rgb(251.92 255 240.83)'))

        # If source channel is 1 resultant channel will be 1
        self.assertColorEqual(Color('white').compose(REDISH, blend='color-dodge'), Color('white'))

    def test_color_burn(self):
        """Test color burn."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='color-burn'), Color('black'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='color-burn'), Color('white'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='color-burn'), Color('rgb(244.88 71.066 0)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='color-burn'), Color('rgb(145.71 6.407 145.25)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='color-burn'), Color('rgb(0 198.62 0)'))

        # If source is  channel is 0, resultant channel will be 0
        self.assertColorEqual(Color('black').compose(REDISH, blend='color-burn'), Color('black'))

    def test_difference(self):
        """Test difference."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='difference'), Color('rgb(252 61 153)'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='difference'), Color('rgb(3 194 102)'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='difference'), Color('rgb(7 150 136)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='difference'), Color('rgb(245 138 84)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='difference'), Color('rgb(238 12 220)'))

    def test_exclusion(self):
        """Test exclusion."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='exclusion'), Color('rgb(252 61 153)'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='exclusion'), Color('rgb(3 194 102)'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='exclusion'), Color('rgb(12.765 171.05 149.6)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='exclusion'), Color('rgb(245.16 164.79 105.6)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='exclusion'), Color('rgb(238.55 80.675 222.4)'))

    def test_color_hard_light(self):
        """Test hard light."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='hard-light'), Color('rgb(249 0 51)'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='hard-light'), Color('rgb(255 122 255)'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='hard-light'), Color('rgb(254.76 100.95 64.6)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='hard-light'), Color('rgb(13.835 169.79 240.6)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='hard-light'), Color('rgb(13.451 235.67 221.4)'))

    def test_color_soft_light(self):
        """Test soft light."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='soft-light'), Color('black'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='soft-light'), Color('white'))
        self.assertColorEqual(
            Color(REDISH).compose(YELLOWISH, blend='soft-light'),
            Color('rgb(249.83 192.01 24.722)')
        )
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='soft-light'), Color('rgb(249.2 96.747 191.24)'))
        self.assertColorEqual(
            Color(BLUISH).compose(YELLOWISH, blend='soft-light'),
            Color('rgb(235.92 222.75 50.158)')
        )

    def test_hue(self):
        """Test hue."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='hue'), Color('black'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='hue'), Color('white'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='hue'), Color('rgb(255 168.4 210.11)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='hue'), Color('rgb(13.23 172.67 204.23)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='hue'), Color('rgb(113.71 231.66 255)'))

        # sRGB must be forced
        self.assertColorEqual(
            Color(BLUISH).compose(YELLOWISH, blend='hue', space="display-p3"),
            Color(BLUISH).compose(YELLOWISH, blend='hue', space="srgb")
        )

    def test_saturation(self):
        """Test hue."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='saturation'), Color('black'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='saturation'), Color('white'))
        self.assertColorEqual(
            Color(REDISH).compose(YELLOWISH, blend='saturation'),
            Color('rgb(237.54 209.06 46.543)')
        )
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='saturation'), Color('rgb(255 59.357 153.59)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='saturation'), Color('rgb(245.4 211.1 15.403)'))

        # sRGB must be forced
        self.assertColorEqual(
            Color(BLUISH).compose(YELLOWISH, blend='saturation', space="display-p3"),
            Color(BLUISH).compose(YELLOWISH, blend='saturation', space="srgb")
        )

    def test_luminosity(self):
        """Test luminosity."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='luminosity'), Color('rgb(128.6 128.6 128.6)'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='luminosity'), Color('rgb(128.6 128.6 128.6)'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='luminosity'), Color('rgb(161.06 137.04 0)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='luminosity'), Color('rgb(255 86.175 167.49)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='luminosity'), Color('rgb(182.76 155.5 0)'))

        # sRGB must be forced
        self.assertColorEqual(
            Color(BLUISH).compose(YELLOWISH, blend='luminosity', space="display-p3"),
            Color(BLUISH).compose(YELLOWISH, blend='luminosity', space="srgb")
        )

    def test_color(self):
        """Test color."""

        self.assertColorEqual(Color(REDISH).compose('black', blend='color'), Color('black'))
        self.assertColorEqual(Color(REDISH).compose('white', blend='color'), Color('white'))
        self.assertColorEqual(Color(REDISH).compose(YELLOWISH, blend='color'), Color('rgb(255 168.4 210.11)'))
        self.assertColorEqual(Color(BLUISH).compose(REDISH, blend='color'), Color('rgb(0 177.73 212.9)'))
        self.assertColorEqual(Color(BLUISH).compose(YELLOWISH, blend='color'), Color('rgb(113.71 231.66 255)'))

        # sRGB must be forced
        self.assertColorEqual(
            Color(BLUISH).compose(YELLOWISH, blend='color', space="display-p3"),
            Color(BLUISH).compose(YELLOWISH, blend='color', space="srgb")
        )



"""Chromatic adaptation transforms."""
from . import util
from .util import Matrix, MutableMatrix, Vector, MutableVector
from . spaces import WHITES
from functools import lru_cache
from typing import Tuple, Dict, cast

# Conversion matrices
CATS = {
    "bradford": [
        # http://brucelindbloom.com/Eqn_ChromAdapt.html
        # https://hrcak.srce.hr/file/95370
        [0.8951000, 0.2664000, -0.1614000],
        [-0.7502000, 1.7135000, 0.0367000],
        [0.0389000, -0.0685000, 1.0296000]
    ],
    "von-kries": [
        # http://brucelindbloom.com/Eqn_ChromAdapt.html
        # https://hrcak.srce.hr/file/95370
        [0.4002400, 0.7076000, -0.0808100],
        [-0.2263000, 1.1653200, 0.0457000],
        [0.0000000, 0.0000000, 0.9182200]
    ],
    "xyz-scaling": [
        # http://brucelindbloom.com/Eqn_ChromAdapt.html
        # https://hrcak.srce.hr/file/95370
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]
    ],
    "cat02": [
        # https://en.wikipedia.org/wiki/CIECAM02#CAT02
        [0.7328000, 0.4296000, -0.1624000],
        [-0.7036000, 1.6975000, 0.0061000],
        [0.0030000, 0.0136000, 0.9834000]
    ],
    "cmccat97": [
        # https://hrcak.srce.hr/file/95370
        [0.8951000, -0.7502000, 0.0389000],
        [0.2664000, 1.7135000, 0.0685000],
        [-0.1614000, 0.0367000, 1.0296000],
    ],
    "sharp": [
        # https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.14.918&rep=rep1&type=pdf
        [1.2694000, -0.0988000, -0.1706000],
        [-0.8364000, 1.8006000, 0.0357000],
        [0.0297000, -0.0315000, 1.0018000]
    ],
    'cmccat2000': [
        # https://hrcak.srce.hr/file/95370
        # https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.14.918&rep=rep1&type=pdf
        [0.7982000, 0.3389000, -0.1371000],
        [-0.5918000, 1.5512000, 0.0406000],
        [0.0008000, 0.0239000, 0.9753000]
    ],
    "cat16": [
        # https://arxiv.org/pdf/1802.06067.pdf
        [-0.401288, -0.250268, -0.002079],
        [0.650173, 1.204414, 0.048952],
        [-0.051461, -0.045854, -0.953127]
    ]
}  # type: Dict[str, Matrix]


@lru_cache(maxsize=20)
def calc_adaptation_matrices(
    w1: str,
    w2: str,
    method: str = 'bradford'
) -> Tuple[MutableMatrix, MutableMatrix]:
    """
    Get the von Kries based adaptation matrix based on the method and illuminants.

    Since these calculated matrices are cached, this greatly reduces
    performance hit as the initial matrices only have to be calculated
    once for a given pair of white points and CAT.

    Granted, we are currently, capped at 20 in the cache, but the average user
    isn't going to be swapping between over 20 methods and white points in a
    short period of time. We could always increase the cache if necessary.
    """

    try:
        m = CATS[method]
    except KeyError:  # pragma: no cover
        raise ValueError('Unknown chromatic adaptation method encountered: {}'.format(method))
    mi = util.inv(m)

    try:
        first = util.dot(m, util.xy_to_xyz(WHITES[w1]))
    except KeyError:  # pragma: no cover
        raise ValueError('Unknown white point encountered: {}'.format(w1))

    try:
        second = util.dot(m, util.xy_to_xyz(WHITES[w2]))
    except KeyError:  # pragma: no cover
        raise ValueError('Unknown white point encountered: {}'.format(w2))

    m2 = util.diag(cast(Vector, util.divide(cast(Vector, first), cast(Vector, second))))
    adapt = util.dot(mi, util.dot(m2, m))

    return cast(MutableMatrix, adapt), util.inv(cast(Matrix, adapt))


def get_adaptation_matrix(w1: str, w2: str, method: str) -> MutableMatrix:
    """
    Get the appropriate matrix for chromatic adaptation.

    If the required matrices are not in the cache, they will be calculated.
    Since white points are sorted by name, regardless of the requested
    conversion direction, the same matrices will be retrieved from the cache.
    """

    a, b = sorted([w1, w2])
    m, mi = calc_adaptation_matrices(a, b, method)
    return mi if a != w2 else m


def chromatic_adaptation(w1: str, w2: str, xyz: Vector, method: str = 'bradford') -> MutableVector:
    """Chromatic adaptation."""

    if w1 == w2:
        # No adaptation is needed if the white points are identical.
        return list(xyz)
    else:
        # Get the appropriate chromatic adaptation matrix and apply.
        return cast(MutableVector, util.dot(get_adaptation_matrix(w1, w2, method), xyz))



"""Lch class."""
from ...spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, Lchish, FLG_ANGLE, FLG_PERCENT
from ... import util
import re
import math
from ...util import MutableVector
from typing import Tuple

ACHROMATIC_THRESHOLD = 0.0000000002


def lab_to_lch(lab: MutableVector) -> MutableVector:
    """Lab to Lch."""

    l, a, b = lab

    c = math.sqrt(a ** 2 + b ** 2)
    h = math.degrees(math.atan2(b, a))

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if c < ACHROMATIC_THRESHOLD:
        h = util.NaN

    test = [l, c, util.constrain_hue(h)]
    return test


def lch_to_lab(lch: MutableVector) -> MutableVector:
    """Lch to Lab."""

    l, c, h = lch
    h = util.no_nan(h)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if c < 0.0:
        c = 0.0

    return [
        l,
        c * math.cos(math.radians(h)),
        c * math.sin(math.radians(h))
    ]


class Lch(Lchish, Space):
    """Lch class."""

    BASE = "lab"
    NAME = "lch"
    SERIALIZE = ("--lch",)
    CHANNEL_NAMES = ("l", "c", "h")
    CHANNEL_ALIASES = {
        "lightness": "l",
        "chroma": "c",
        "hue": "h"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D50"
    BOUNDS = (
        # I think chroma, specifically should be clamped.
        # Some libraries don't to prevent rounding issues. We should only get
        # negative chroma via direct user input, but when translating to
        # Lab, this will be corrected.
        GamutUnbound(0.0, 100.0, FLG_PERCENT),
        GamutUnbound(0.0, 100.0),
        GamutUnbound(0.0, 360.0, FLG_ANGLE)
    )

    @property
    def l(self) -> float:
        """Lightness."""

        return self._coords[0]

    @l.setter
    def l(self, value: float) -> None:
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def c(self) -> float:
        """Chroma."""

        return self._coords[1]

    @c.setter
    def c(self, value: float) -> None:
        """chroma."""

        self._coords[1] = self._handle_input(value)

    @property
    def h(self) -> float:
        """Hue."""

        return self._coords[2]

    @h.setter
    def h(self, value: float) -> None:
        """Shift the hue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords: MutableVector, alpha: float) -> Tuple[MutableVector, float]:
        """On color update."""

        if coords[1] < ACHROMATIC_THRESHOLD:
            coords[2] = util.NaN
        return coords, alpha

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To Lab from Lch."""

        return lch_to_lab(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From Lab to Lch."""

        return lab_to_lch(coords)



"""Display-p3 color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb import SRGB, lin_srgb, gam_srgb
from .. import util
import re
from ..util import MutableVector
from typing import cast

RGB_TO_XYZ = [
    [4.8657094864821610e-01, 2.6566769316909306e-01, 1.9821728523436244e-01],
    [2.2897456406974875e-01, 6.9173852183650641e-01, 7.9286914093744970e-02],
    [-3.9720755169334868e-17, 4.5113381858902638e-02, 1.0439443689009755e+00]
]

XYZ_TO_RGB = [
    [2.4934969119414254, -0.9313836179191239, -0.40271078445071695],
    [-0.8294889695615746, 1.7626640603183463, 0.023624685841943566],
    [0.03584583024378447, -0.07617238926804185, 0.9568845240076876]
]


def lin_p3_to_xyz(rgb: MutableVector) -> MutableVector:
    """
    Convert an array of linear-light image-p3 values to CIE XYZ using  D65 (no chromatic adaptation).

    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    # 0 was computed as -3.972075516933488e-17
    return cast(MutableVector, util.dot(RGB_TO_XYZ, rgb))


def xyz_to_lin_p3(xyz: MutableVector) -> MutableVector:
    """Convert XYZ to linear-light P3."""

    return cast(MutableVector, util.dot(XYZ_TO_RGB, xyz))


def lin_p3(rgb: MutableVector) -> MutableVector:
    """Convert an array of image-p3 RGB values in the range 0.0 - 1.0 to linear light (un-corrected) form."""

    return lin_srgb(rgb)  # same as sRGB


def gam_p3(rgb: MutableVector) -> MutableVector:
    """Convert an array of linear-light image-p3 RGB  in the range 0.0-1.0 to gamma corrected form."""

    return gam_srgb(rgb)  # same as sRGB


class DisplayP3(SRGB):
    """Display-p3 class."""

    BASE = "xyz-d65"
    NAME = "display-p3"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=NAME, channels=3))
    WHITE = "D65"

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ from Display P3."""

        return lin_p3_to_xyz(lin_p3(coords))

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ to Display P3."""

        return gam_p3(xyz_to_lin_p3(coords))



"""A98 RGB color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb import SRGB
from .. import util
import re
from ..util import MutableVector
from typing import cast

RGB_TO_XYZ = [
    [0.5766690429101304, 0.18555823790654635, 0.18822864623499475],
    [0.297344975250536, 0.6273635662554663, 0.0752914584939979],
    [0.027031361386412336, 0.07068885253582725, 0.9913375368376391]
]

XYZ_TO_RGB = [
    [2.041587903810747, -0.5650069742788599, -0.34473135077832967],
    [-0.9692436362808794, 1.8759675015077197, 0.04155505740717558],
    [0.013444280632031149, -0.11836239223101835, 1.0151749943912052]
]


def lin_a98rgb_to_xyz(rgb: MutableVector) -> MutableVector:
    """
    Convert an array of linear-light a98-rgb values to CIE XYZ using D50.D65.

    (so no chromatic adaptation needed afterwards)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    which has greater numerical precision than section 4.3.5.3 of
    https://www.adobe.com/digitalimag/pdfs/AdobeRGB1998.pdf
    """

    return cast(MutableVector, util.dot(RGB_TO_XYZ, rgb))


def xyz_to_lin_a98rgb(xyz: MutableVector) -> MutableVector:
    """Convert XYZ to linear-light a98-rgb."""

    return cast(MutableVector, util.dot(XYZ_TO_RGB, xyz))


def lin_a98rgb(rgb: MutableVector) -> MutableVector:
    """Convert an array of a98-rgb values in the range 0.0 - 1.0 to linear light (un-corrected) form."""

    return [util.npow(val, 563 / 256) for val in rgb]


def gam_a98rgb(rgb: MutableVector) -> MutableVector:
    """Convert an array of linear-light a98-rgb  in the range 0.0-1.0 to gamma corrected form."""

    return [util.npow(val, 256 / 563) for val in rgb]


class A98RGB(SRGB):
    """A98 RGB class."""

    BASE = "xyz-d65"
    NAME = "a98-rgb"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=NAME, channels=3))
    WHITE = "D65"

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ from A98 RGB."""

        return lin_a98rgb_to_xyz(lin_a98rgb(coords))

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ to A98 RGB."""

        return gam_a98rgb(xyz_to_lin_a98rgb(coords))



"""XYZ class."""
from ..spaces import RE_DEFAULT_MATCH
from .xyz_d65 import XYZD65
import re


class XYZD50(XYZD65):
    """XYZ D50 class."""

    BASE = "xyz-d65"
    NAME = "xyz-d50"
    SERIALIZE = ("xyz-d50",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D50"



"""
Okhsl class.

Adapted to ColorAide Python and ColorAide by Isaac Muse (2021)

---- License ----

Copyright (c) 2021 Björn Ottosson

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..spaces import Space, RE_DEFAULT_MATCH, FLG_ANGLE, FLG_OPT_PERCENT, GamutBound, Cylindrical
from .oklab import oklab_to_linear_srgb
from .. import util
import re
import math
import sys
from ..util import MutableVector
from typing import Tuple, Optional

FLT_MAX = sys.float_info.max

K_1 = 0.206
K_2 = 0.03
K_3 = (1.0 + K_1) / (1.0 + K_2)


def toe(x: float) -> float:
    """Toe function for L_r."""

    return 0.5 * (K_3 * x - K_1 + math.sqrt((K_3 * x - K_1) * (K_3 * x - K_1) + 4 * K_2 * K_3 * x))


def toe_inv(x: float) -> float:
    """Inverse toe function for L_r."""

    return (x ** 2 + K_1 * x) / (K_3 * (x + K_2))


def to_st(cusp: MutableVector) -> MutableVector:
    """To ST."""

    l, c = cusp
    return [c / l, c / (1 - l)]


def get_st_mid(a: float, b: float) -> MutableVector:
    """
    Returns a smooth approximation of the location of the cusp.

    This polynomial was created by an optimization process.
    It has been designed so that S_mid < S_max and T_mid < T_max.
    """

    s = 0.11516993 + 1.0 / (
        7.44778970 + 4.15901240 * b +
        a * (
            -2.19557347 + 1.75198401 * b +
            a * (
                -2.13704948 - 10.02301043 * b +
                a * (
                    -4.24894561 + 5.38770819 * b + 4.69891013 * a
                )
            )
        )
    )

    t = 0.11239642 + 1.0 / (
        1.61320320 - 0.68124379 * b +
        a * (
            0.40370612 + 0.90148123 * b +
            a * (
                -0.27087943 + 0.61223990 * b +
                a * (
                    0.00299215 - 0.45399568 * b - 0.14661872 * a
                )
            )
        )
    )

    return [s, t]


def find_cusp(a: float, b: float) -> MutableVector:
    """
    Finds L_cusp and C_cusp for a given hue.

    `a` and `b` must be normalized so `a^2 + b^2 == 1`.
    """

    # First, find the maximum saturation (saturation `S = C/L`)
    s_cusp = compute_max_saturation(a, b)

    # Convert to linear sRGB to find the first point where at least one of r, g or b >= 1:
    r, g, b = oklab_to_linear_srgb([1, s_cusp * a, s_cusp * b])
    l_cusp = util.nth_root(1.0 / max(max(r, g), b), 3)
    c_cusp = l_cusp * s_cusp

    return [l_cusp, c_cusp]


def find_gamut_intersection(
    a: float,
    b: float,
    l1: float,
    c1: float,
    l0: float,
    cusp: Optional[MutableVector] = None
) -> float:
    """
    Finds intersection of the line.

    Defined by the following:

    ```
    L = L0 * (1 - t) + t * L1
    C = t * C1
    ```

    `a` and `b` must be normalized so `a^2 + b^2 == 1`.
    """

    if cusp is None:  # pragma: no cover
        cusp = find_cusp(a, b)

    # Find the intersection for upper and lower half separately
    if ((l1 - l0) * cusp[1] - (cusp[0] - l0) * c1) <= 0.0:
        # Lower half
        t = cusp[1] * l0 / (c1 * cusp[0] + cusp[1] * (l0 - l1))
    else:
        # Upper half

        # First intersect with triangle
        t = cusp[1] * (l0 - 1.0) / (c1 * (cusp[0] - 1.0) + cusp[1] * (l0 - l1))

        # Then one step Halley's method
        dl = l1 - l0
        dc = c1

        k_l = +0.3963377774 * a + 0.2158037573 * b
        k_m = -0.1055613458 * a - 0.0638541728 * b
        k_s = -0.0894841775 * a - 1.2914855480 * b

        l_dt = dl + dc * k_l
        m_dt = dl + dc * k_m
        s_dt = dl + dc * k_s

        # If higher accuracy is required, 2 or 3 iterations of the following block can be used:
        L = l0 * (1.0 - t) + t * l1
        C = t * c1

        l_ = L + C * k_l
        m_ = L + C * k_m
        s_ = L + C * k_s

        l = l_ ** 3
        m = m_ ** 3
        s = s_ ** 3

        ldt = 3 * l_dt * (l_ ** 2)
        mdt = 3 * m_dt * (m_ ** 2)
        sdt = 3 * s_dt * (s_ ** 2)

        ldt2 = 6 * (l_dt ** 2) * l_
        mdt2 = 6 * (m_dt ** 2) * m_
        sdt2 = 6 * (s_dt ** 2) * s_

        r = 4.0767416621 * l - 3.3077115913 * m + 0.2309699292 * s - 1
        r1 = 4.0767416621 * ldt - 3.3077115913 * mdt + 0.2309699292 * sdt
        r2 = 4.0767416621 * ldt2 - 3.3077115913 * mdt2 + 0.2309699292 * sdt2

        u_r = r1 / (r1 * r1 - 0.5 * r * r2)
        t_r = -r * u_r

        g = -1.2684380046 * l + 2.6097574011 * m - 0.3413193965 * s - 1
        g1 = -1.2684380046 * ldt + 2.6097574011 * mdt - 0.3413193965 * sdt
        g2 = -1.2684380046 * ldt2 + 2.6097574011 * mdt2 - 0.3413193965 * sdt2

        u_g = g1 / (g1 * g1 - 0.5 * g * g2)
        t_g = -g * u_g

        b = -0.0041960863 * l - 0.7034186147 * m + 1.7076147010 * s - 1
        b1 = -0.0041960863 * ldt - 0.7034186147 * mdt + 1.7076147010 * sdt
        b2 = -0.0041960863 * ldt2 - 0.7034186147 * mdt2 + 1.7076147010 * sdt2

        u_b = b1 / (b1 * b1 - 0.5 * b * b2)
        t_b = -b * u_b

        t_r = t_r if u_r >= 0.0 else FLT_MAX
        t_g = t_g if u_g >= 0.0 else FLT_MAX
        t_b = t_b if u_b >= 0.0 else FLT_MAX

        t += min(t_r, min(t_g, t_b))

    return t


def get_cs(lab: MutableVector) -> MutableVector:
    """Get Cs."""

    l, a, b = lab

    cusp = find_cusp(a, b)

    c_max = find_gamut_intersection(a, b, l, 1, l, cusp)
    st_max = to_st(cusp)

    # Scale factor to compensate for the curved part of gamut shape:
    k = c_max / min((l * st_max[0]), (1 - l) * st_max[1])

    st_mid = get_st_mid(a, b)

    # Use a soft minimum function, instead of a sharp triangle shape to get a smooth value for chroma.
    c_a = l * st_mid[0]
    c_b = (1.0 - l) * st_mid[1]
    c_mid = 0.9 * k * math.sqrt(math.sqrt(1.0 / (1.0 / (c_a ** 4) + 1.0 / (c_b ** 4))))

    # For `C_0`, the shape is independent of hue, so `ST` are constant.
    # Values picked to roughly be the average values of `ST`.
    c_a = l * 0.4
    c_b = (1.0 - l) * 0.8

    # Use a soft minimum function, instead of a sharp triangle shape to get a smooth value for chroma.
    c_0 = math.sqrt(1.0 / (1.0 / (c_a ** 2) + 1.0 / (c_b ** 2)))

    return [c_0, c_mid, c_max]


def compute_max_saturation(a: float, b: float) -> float:
    """
    Finds the maximum saturation possible for a given hue that fits in sRGB.

    Saturation here is defined as `S = C/L`.
    `a` and `b` must be normalized so `a^2 + b^2 == 1`.
    """

    # Max saturation will be when one of r, g or b goes below zero.

    # Select different coefficients depending on which component goes below zero first.

    if (-1.88170328 * a - 0.80936493 * b) > 1:
        # Red component
        k0 = 1.19086277
        k1 = 1.76576728
        k2 = 0.59662641
        k3 = 0.75515197
        k4 = 0.56771245
        wl = 4.0767416621
        wm = -3.3077115913
        ws = 0.2309699292

    elif (1.81444104 * a - 1.19445276 * b) > 1:
        # Green component
        k0 = 0.73956515
        k1 = -0.45954404
        k2 = 0.08285427
        k3 = 0.12541070
        k4 = 0.14503204
        wl = -1.2684380046
        wm = 2.6097574011
        ws = -0.3413193965

    else:
        # Blue component
        k0 = 1.35733652
        k1 = -0.00915799
        k2 = -1.15130210
        k3 = -0.50559606
        k4 = 0.00692167
        wl = -0.0041960863
        wm = -0.7034186147
        ws = 1.7076147010

    # Approximate max saturation using a polynomial:
    sat = k0 + k1 * a + k2 * b + k3 * (a ** 2) + k4 * a * b

    # Do one step Halley's method to get closer.
    # This gives an error less than 10e6, except for some blue hues where the `dS/dh` is close to infinite.
    # This should be sufficient for most applications, otherwise do two/three steps.

    k_l = 0.3963377774 * a + 0.2158037573 * b
    k_m = -0.1055613458 * a - 0.0638541728 * b
    k_s = -0.0894841775 * a - 1.2914855480 * b

    l_ = 1.0 + sat * k_l
    m_ = 1.0 + sat * k_m
    s_ = 1.0 + sat * k_s

    l = l_ ** 3
    m = m_ ** 3
    s = s_ ** 3

    l_ds = 3.0 * k_l * (l_ ** 2)
    m_ds = 3.0 * k_m * (m_ ** 2)
    s_ds = 3.0 * k_s * (s_ ** 2)

    l_ds2 = 6.0 * (k_l ** 2) * l_
    m_ds2 = 6.0 * (k_m ** 2) * m_
    s_ds2 = 6.0 * (k_s ** 2) * s_

    f = wl * l + wm * m + ws * s
    f1 = wl * l_ds + wm * m_ds + ws * s_ds
    f2 = wl * l_ds2 + wm * m_ds2 + ws * s_ds2

    sat = sat - f * f1 / ((f1 ** 2) - 0.5 * f * f2)

    return sat


def okhsl_to_oklab(hsl: MutableVector) -> MutableVector:
    """Convert Okhsl to sRGB."""

    h, s, l = hsl
    h = util.no_nan(h) / 360.0

    L = toe_inv(l)
    a = b = 0.0

    if L != 0 and L != 1 and s != 0:
        a_ = math.cos(2.0 * math.pi * h)
        b_ = math.sin(2.0 * math.pi * h)

        c_0, c_mid, c_max = get_cs([L, a_, b_])

        # Interpolate the three values for C so that:
        # ```
        # At s=0: dC/ds = C_0, C=0
        # At s=0.8: C=C_mid
        # At s=1.0: C=C_max
        # ```

        mid = 0.8
        mid_inv = 1.25

        if s < mid:
            t = mid_inv * s
            k_0 = 0.0
            k_1 = mid * c_0
            k_2 = (1.0 - k_1 / c_mid)

        else:
            t = 5 * (s - 0.8)
            k_0 = c_mid
            k_1 = 0.2 * (c_mid ** 2) * (1.25 ** 2) / c_0
            k_2 = 1.0 - k_1 / (c_max - c_mid)

        c = k_0 + t * k_1 / (1.0 - k_2 * t)

        a = c * a_
        b = c * b_

    return [L, a, b]


def oklab_to_okhsl(lab: MutableVector) -> MutableVector:
    """Oklab to Okhsl."""

    c = math.sqrt(lab[1] ** 2 + lab[2] ** 2)

    h = util.NaN
    L = lab[0]
    s = 0.0

    if c != 0 and L != 0:
        a_ = lab[1] / c
        b_ = lab[2] / c

        h = 0.5 + 0.5 * math.atan2(-lab[2], -lab[1]) / math.pi

        c_0, c_mid, c_max = get_cs([L, a_, b_])

        # Inverse of the interpolation in `okhsl_to_srgb`:

        mid = 0.8
        mid_inv = 1.25

        if (c < c_mid):
            k_1 = mid * c_0
            k_2 = 1.0 - k_1 / c_mid

            t = c / (k_1 + k_2 * c)
            s = t * mid

        else:
            k_0 = c_mid
            k_1 = 0.2 * (c_mid ** 2) * (mid_inv ** 2) / c_0
            k_2 = (1.0 - (k_1) / (c_max - c_mid))

            t = (c - k_0) / (k_1 + k_2 * (c - k_0))
            s = mid + 0.2 * t

    l = toe(L)

    if s == 0:
        h = util.NaN

    return [util.constrain_hue(h * 360), s, l]


class Okhsl(Cylindrical, Space):
    """HSL class."""

    BASE = "oklab"
    NAME = "okhsl"
    SERIALIZE = ("--okhsl",)
    CHANNEL_NAMES = ("h", "s", "l")
    CHANNEL_ALIASES = {
        "hue": "h",
        "saturation": "s",
        "lightness": "l"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"
    GAMUT_CHECK = "srgb"

    BOUNDS = (
        GamutBound(0.0, 360.0, FLG_ANGLE),
        GamutBound(0.0, 1.0, FLG_OPT_PERCENT),
        GamutBound(0.0, 1.0, FLG_OPT_PERCENT)
    )

    @property
    def h(self) -> float:
        """Hue channel."""

        return self._coords[0]

    @h.setter
    def h(self, value: float) -> None:
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def s(self) -> float:
        """Saturation channel."""

        return self._coords[1]

    @s.setter
    def s(self, value: float) -> None:
        """Saturate or unsaturate the color by the given factor."""

        self._coords[1] = self._handle_input(value)

    @property
    def l(self) -> float:
        """Lightness channel."""

        return self._coords[2]

    @l.setter
    def l(self, value: float) -> None:
        """Set lightness channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords: MutableVector, alpha: float) -> Tuple[MutableVector, float]:
        """On color update."""

        if coords[1] == 0:
            coords[0] = util.NaN
        return coords, alpha

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To Oklab from Okhsl."""

        return okhsl_to_oklab(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From Oklab to Okhsl."""

        return oklab_to_okhsl(coords)



"""Lab class."""
from ...spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, FLG_PERCENT, Labish
from ... import util
import re
from ...util import Vector, MutableVector
from typing import cast

EPSILON = 216 / 24389  # `6^3 / 29^3`
EPSILON3 = 6 / 29  # Cube root of EPSILON
KAPPA = 24389 / 27
KE = 8  # KAPPA * EPSILON = 8


def lab_to_xyz(lab: MutableVector, white: Vector) -> MutableVector:
    """
    Convert Lab to D50-adapted XYZ.

    http://www.brucelindbloom.com/Eqn_Lab_to_XYZ.html

    While the derivation is different than the specification, the results are the same as Appendix D:
    https://www.cdvplus.cz/file/3-publikace-cie15-2004/
    """

    l, a, b = lab

    # compute `f`, starting with the luminance-related term
    fy = (l + 16) / 116
    fx = a / 500 + fy
    fz = fy - b / 200

    # compute `xyz`
    xyz = [
        fx ** 3 if fx > EPSILON3 else (116 * fx - 16) / KAPPA,
        fy ** 3 if l > KE else l / KAPPA,
        fz ** 3 if fz > EPSILON3 else (116 * fz - 16) / KAPPA
    ]

    # Compute XYZ by scaling `xyz` by reference `white`
    return cast(MutableVector, util.multiply(xyz, util.xy_to_xyz(white)))


def xyz_to_lab(xyz: MutableVector, white: Vector) -> MutableVector:
    """
    Assuming XYZ is relative to D50, convert to CIE Lab from CIE standard.

    http://www.brucelindbloom.com/Eqn_XYZ_to_Lab.html

    While the derivation is different than the specification, the results are the same:
    https://www.cdvplus.cz/file/3-publikace-cie15-2004/
    """

    # compute `xyz`, which is XYZ scaled relative to reference white
    xyz = cast(MutableVector, util.divide(xyz, util.xy_to_xyz(white)))
    # Compute `fx`, `fy`, and `fz`
    fx, fy, fz = [util.cbrt(i) if i > EPSILON else (KAPPA * i + 16) / 116 for i in xyz]

    return [
        (116.0 * fy) - 16.0,
        500.0 * (fx - fy),
        200.0 * (fy - fz)
    ]


class Lab(Labish, Space):
    """Lab class."""

    BASE = "xyz-d50"
    NAME = "lab"
    SERIALIZE = ("--lab",)
    CHANNEL_NAMES = ("l", "a", "b")
    CHANNEL_ALIASES = {
        "lightness": "l"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D50"
    BOUNDS = (
        GamutUnbound(0.0, 100.0, FLG_PERCENT),  # Technically we could/should clamp the zero side.
        GamutUnbound(-160, 160),  # No limit, but we could impose one +/-160?
        GamutUnbound(-160, 160)  # No limit, but we could impose one +/-160?
    )

    @property
    def l(self) -> float:
        """L channel."""

        return self._coords[0]

    @l.setter
    def l(self, value: float) -> None:
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def a(self) -> float:
        """A channel."""

        return self._coords[1]

    @a.setter
    def a(self, value: float) -> None:
        """A axis."""

        self._coords[1] = self._handle_input(value)

    @property
    def b(self) -> float:
        """B channel."""

        return self._coords[2]

    @b.setter
    def b(self, value: float) -> None:
        """B axis."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ D50 from Lab."""

        return lab_to_xyz(coords, cls.white())

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ D50 to Lab."""

        return xyz_to_lab(coords, cls.white())



"""HSL class."""
from ...spaces import Space, RE_DEFAULT_MATCH, FLG_ANGLE, FLG_OPT_PERCENT, GamutBound, Cylindrical
from ... import util
import re
from ...util import MutableVector
from typing import Tuple


def srgb_to_hsl(rgb: MutableVector) -> MutableVector:
    """SRGB to HSL."""

    r, g, b = rgb
    mx = max(rgb)
    mn = min(rgb)
    h = util.NaN
    s = 0.0
    l = (mn + mx) / 2
    c = mx - mn

    if c != 0.0:
        if mx == r:
            h = (g - b) / c
        elif mx == g:
            h = (b - r) / c + 2.0
        else:
            h = (r - g) / c + 4.0
        s = 0 if l == 0 or l == 1 else (mx - l) / min(l, 1 - l)
        h *= 60.0
        if s == 0:
            h = util.NaN

    return [util.constrain_hue(h), s, l]


def hsl_to_srgb(hsl: MutableVector) -> MutableVector:
    """
    HSL to RGB.

    https://en.wikipedia.org/wiki/HSL_and_HSV#HSL_to_RGB_alternative
    """

    h, s, l = hsl
    h = util.no_nan(h)
    h = h % 360

    def f(n: int) -> float:
        """Calculate the channels."""
        k = (n + h / 30) % 12
        a = s * min(l, 1 - l)
        return l - a * max(-1, min(k - 3, 9 - k, 1))

    return [f(0), f(8), f(4)]


class HSL(Cylindrical, Space):
    """HSL class."""

    BASE = "srgb"
    NAME = "hsl"
    SERIALIZE = ("--hsl",)
    CHANNEL_NAMES = ("h", "s", "l")
    CHANNEL_ALIASES = {
        "hue": "h",
        "saturation": "s",
        "lightness": "l"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"
    GAMUT_CHECK = "srgb"

    BOUNDS = (
        GamutBound(0.0, 360.0, FLG_ANGLE),
        GamutBound(0.0, 1.0, FLG_OPT_PERCENT),
        GamutBound(0.0, 1.0, FLG_OPT_PERCENT)
    )

    @property
    def h(self) -> float:
        """Hue channel."""

        return self._coords[0]

    @h.setter
    def h(self, value: float) -> None:
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def s(self) -> float:
        """Saturation channel."""

        return self._coords[1]

    @s.setter
    def s(self, value: float) -> None:
        """Saturate or unsaturate the color by the given factor."""

        self._coords[1] = self._handle_input(value)

    @property
    def l(self) -> float:
        """Lightness channel."""

        return self._coords[2]

    @l.setter
    def l(self, value: float) -> None:
        """Set lightness channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords: MutableVector, alpha: float) -> Tuple[MutableVector, float]:
        """On color update."""

        if coords[1] == 0:
            coords[0] = util.NaN

        return coords, alpha

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To sRGB from HSL."""

        return hsl_to_srgb(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From sRGB to HSL."""

        return srgb_to_hsl(coords)



"""ColorAide Library."""
from .__meta__ import __version_info__, __version__  # noqa: F401
from .color import Color, ColorMatch
from .interpolate import Piecewise, Lerp
from .util import NaN

__all__ = ("Color", "ColorMatch", "NaN", "Piecewise", "Lerp")



"""Test LCH library."""
import unittest
import math
from . import util
from coloraide import Color, NaN


class TestLCHInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test LCH."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("lch", [20, 10, 130]), Color('lch(20% 10 130)'))

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("lch", [20, 10, 130])), Color('lch(20% 10 130)'))

    def test_color(self):
        """Test color input/output format."""

        args = {"color": True}
        color = "color(--lch 20% 10 130)"

        self.assertEqual(Color(color).to_string(**args), 'color(--lch 20 10 130)')

        color = "color(--lch 20% 10 130 / 0.5)"
        self.assertEqual(Color(color).to_string(**args), 'color(--lch 20 10 130 / 0.5)')

        color = "color(--lch 20% 10 130 / 50%)"
        self.assertEqual(Color(color).to_string(**args), 'color(--lch 20 10 130 / 0.5)')

    def test_comma(self):
        """Test comma input and comma output format."""

        args = {"comma": True}

        color = "lch(20%, 10, 130)"
        lch = Color(color)
        self.assertEqual(color, lch.to_string(**args))

        color = "lch(20%, 10, 130, 1)"
        lch = Color(color)
        self.assertEqual("lch(20%, 10, 130)", lch.to_string(**args))

        color = "lch(20%, 10, 130, 0.2)"
        lch = Color(color)
        self.assertEqual("lch(20%, 10, 130, 0.2)", lch.to_string(**args))

    def test_space(self):
        """Test space input and space output format."""

        args = {}

        color = "lch(20% 10 130)"
        lch = Color(color)
        self.assertEqual(color, lch.to_string(**args))

        color = "lch(20% 10 130 / 1)"
        lch = Color(color)
        self.assertEqual("lch(20% 10 130)", lch.to_string(**args))

        color = "lch(20% 10 130 / 0.2)"
        lch = Color(color)
        self.assertEqual(color, lch.to_string(**args))

    def test_percent(self):
        """Test that percents work properly."""

        args = {"comma": True}

        color = "lch(20%, 10, 130, 100%)"
        lch = Color(color)
        self.assertEqual("lch(20%, 10, 130)", lch.to_string(**args))

        color = "lch(20%, 10, 130, 20%)"
        lch = Color(color)
        self.assertEqual("lch(20%, 10, 130, 0.2)", lch.to_string(**args))

        args["comma"] = False

        color = "lch(20% 10 130 / 100%)"
        lch = Color(color)
        self.assertEqual("lch(20% 10 130)", lch.to_string(**args))

        color = "lch(20% 10 130 / 20%)"
        lch = Color(color)
        self.assertEqual("lch(20% 10 130 / 0.2)", lch.to_string(**args))

    def test_no_alpha(self):
        """Test no alpha."""

        args = {"comma": True, "alpha": False}

        color = "lch(20%, 10, 130, 0.2)"
        lch = Color(color)
        self.assertEqual("lch(20%, 10, 130)", lch.to_string(**args))

        args["comma"] = False

        color = "lch(20% 10 130 / 0.2)"
        lch = Color(color)
        self.assertEqual("lch(20% 10 130)", lch.to_string(**args))

    def test_force_alpha(self):
        """Test force alpha."""

        args = {"comma": True, "alpha": True}

        color = "lch(20%, 10, 130, 1)"
        lch = Color(color)
        self.assertEqual("lch(20%, 10, 130, 1)", lch.to_string(**args))

        args["comma"] = False

        color = "lch(20% 10 130 / 1)"
        lch = Color(color)
        self.assertEqual("lch(20% 10 130 / 1)", lch.to_string(**args))

    def test_precision(self):
        """Test precision."""

        color = 'color(--lch 20.1234567% 10.1234567 130.1234567)'
        self.assertEqual(Color(color).to_string(), 'lch(20.123% 10.123 130.12)')
        self.assertEqual(Color(color).to_string(precision=3), 'lch(20.1% 10.1 130)')
        self.assertEqual(Color(color).to_string(precision=0), 'lch(20% 10 130)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'lch(20.12345669999999842048055143095552921295166015625% 10.1234567000000001968373908312059938907623291015625 130.123456699999991315053193829953670501708984375)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('color(--lch 20% 200 120)').to_string(),
            'lch(20% 200 120)'
        )

        self.assertEqual(
            Color('color(--lch 20% 200 120)').to_string(fit="clip"),
            'lch(20% 200 120)'
        )

        self.assertEqual(
            Color('color(--lch 20% 200 120)').to_string(fit=False),
            'lch(20% 200 120)'
        )

    def test_hue_inputs(self):
        """Test hue inputs."""

        color = "lch(90%, 50, 90deg)"
        lch = Color(color)
        self.assertEqual("lch(90% 50 90)", lch.to_string())

        color = "lch(90%, 50, {:f}rad)".format(math.radians(90))
        lch = Color(color)
        self.assertEqual("lch(90% 50 90)", lch.to_string())

        color = "lch(90%, 50, 100grad)"
        lch = Color(color)
        self.assertEqual("lch(90% 50 90)", lch.to_string())

        color = "lch(90%, 50, 0.25turn)"
        lch = Color(color)
        self.assertEqual("lch(90% 50 90)", lch.to_string())


class TestLCHProperties(util.ColorAsserts, unittest.TestCase):
    """Test LCH."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--lch 90% 50 120 / 1)')
        self.assertEqual(c.lightness, 90)
        c.lightness = 80
        self.assertEqual(c.lightness, 80)

    def test_chroma(self):
        """Test `chroma`."""

        c = Color('color(--lch 90% 50 120 / 1)')
        self.assertEqual(c.chroma, 50)
        c.chroma = 40
        self.assertEqual(c.chroma, 40)

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--lch 90% 50 120 / 1)')
        self.assertEqual(c.hue, 120)
        c.hue = 110
        self.assertEqual(c.hue, 110)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--lch 90% 50 120 / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('lch', [90, 50, NaN], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_auto_null(self):
        """Test auto null."""

        c = Color('lch(90% 0 120 / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_near_zero_null(self):
        """
        Test very near zero null.

        This is a fix up to help give more sane hues
        when chroma is very close to zero.
        """

        c = Color('lch(90% 0.000000000009 120 / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_from_lab(self):
        """Test null from Lab conversion."""

        c1 = Color('lab(90% 0 0)')
        c2 = c1.convert('lch')
        self.assertColorEqual(c2, Color('lch(90% 0 0)'))
        self.assertTrue(c2.is_nan('hue'))

    def test_achromatic_hue(self):
        """Test that all RGB-ish colors convert to LCH with a null hue."""

        for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
            for x in range(0, 256):
                color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
                color2 = color.convert('lch')
                self.assertTrue(color2.is_nan('hue'))


class TestQuirks(util.ColorAsserts, unittest.TestCase):
    """Test quirks."""

    def test_chorma_less_than_zero_convert(self):
        """Test handling of negative chroma when converting to lab."""

        self.assertColorEqual(Color('lch(90% -10 120 / 1)').convert('lab'), Color('lab(90% 0 0)'))



"""SRGB Linear color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb import SRGB
import re
from ..util import MutableVector
from typing import cast
from ..import util


RGB_TO_XYZ = [
    [0.4123907992659593, 0.357584339383878, 0.18048078840183432],
    [0.21263900587151024, 0.715168678767756, 0.07219231536073373],
    [0.01933081871559182, 0.11919477979462598, 0.9505321522496608]
]

XYZ_TO_RGB = [
    [3.2409699419045226, -1.537383177570094, -0.49861076029300355],
    [-0.9692436362808796, 1.8759675015077202, 0.04155505740717562],
    [0.055630079696993635, -0.2039769588889765, 1.0569715142428784]
]


def lin_srgb_to_xyz(rgb: MutableVector) -> MutableVector:
    """
    Convert an array of linear-light sRGB values to CIE XYZ using sRGB's own white.

    D65 (no chromatic adaptation)
    """

    return cast(MutableVector, util.dot(RGB_TO_XYZ, rgb))


def xyz_to_lin_srgb(xyz: MutableVector) -> MutableVector:
    """Convert XYZ to linear-light sRGB."""

    return cast(MutableVector, util.dot(XYZ_TO_RGB, xyz))


class SRGBLinear(SRGB):
    """SRGB linear."""

    BASE = 'xyz-d65'
    NAME = "srgb-linear"
    SERIALIZE = ("srgb-linear",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ from SRGB Linear."""

        return lin_srgb_to_xyz(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ to SRGB Linear."""

        return xyz_to_lin_srgb(coords)
